import os

from pykechain.enums import PropertyType, Multiplicity
from pykechain.exceptions import IllegalArgumentError
from pykechain.utils import temp_chdir

# global variable
__mapping_dictionary = None
__edited_one_many = list()  # type: list


def get_mapping_dictionary(clean=False):
    """
    Get a temporary helper to map some keys to some values. Mainly used in the relocation of parts and models.

    :param clean: (optional) boolean flag to reset the mapping dictionary
    :type clean: bool
    :return: singleton dictionary (persistent in the script) for mapping use.
    """
    global __mapping_dictionary
    if not __mapping_dictionary or clean:
        __mapping_dictionary = dict()
    return __mapping_dictionary


def get_edited_one_many(clean=False):
    """
    Get a temporary helper to help relocating Parts with one to many relationships.

    Only used in the relocation of parts and models.

    :param clean: (optional) boolean flag to reset the mapping dictionary
    :type clean: bool
    :return: singleton dictionary (persistent in the script) for mapping use.
    """
    global __edited_one_many
    if not __edited_one_many or clean:
        __edited_one_many = list()
    return __edited_one_many


def relocate_model(part, target_parent, name=None, include_children=True):
    """
    Move the `Part` model to target parent.

    .. versionadded:: 2.3

    :param part: `Part` object to be moved
    :type part: :class:`Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: Part model.
    :raises IllegalArgumentError: if target_parent is descendant of part
    """
    if target_parent.id in get_illegal_targets(part, include={part.id}):
        raise IllegalArgumentError('cannot relocate part "{}" under target parent "{}", because the target is part of '
                                   'its descendants'.format(part.name, target_parent.name))

    # First, if the user doesn't provide the name, then just use the default "Clone - ..." name
    if not name:
        name = "CLONE - {}".format(part.name)

    # The description cannot be added when creating a model, so edit the model after creation.
    part_desc = part.description
    moved_part_model = target_parent.add_model(name=name, multiplicity=part.multiplicity)
    if part_desc:
        moved_part_model.edit(description=str(part_desc))

    # Map the current part model id with newly created part model Object
    get_mapping_dictionary().update({part.id: moved_part_model})

    # Loop through properties and retrieve their type, description and unit
    list_of_properties_sorted_by_order = part.properties
    list_of_properties_sorted_by_order.sort(key=lambda x: x._json_data['order'])
    for prop in list_of_properties_sorted_by_order:
        prop_type = prop.type
        description = prop.description
        unit = prop.unit

        # For KE-chain 3 (PIM2) we have value_options instead of options.
        if prop._client.match_app_version(label='pim', version='>=3.0.0'):
            options = prop._json_data.get('value_options')
        else:
            options = prop._json_data.get('options')

        # On "Part references" properties, the models referenced also need to be added
        if prop_type == PropertyType.REFERENCES_VALUE:
            referenced_part_ids = [referenced_part.id for referenced_part in prop.value]
            moved_prop = moved_part_model.add_property(
                name=prop.name,
                description=description,
                property_type=prop_type,
                default_value=referenced_part_ids,
                options=options
            )

        # On "Attachment" properties, attachments needs to be downloaded and re-uploaded to the new property.
        elif prop_type == PropertyType.ATTACHMENT_VALUE:
            moved_prop = moved_part_model.add_property(
                name=prop.name,
                description=description,
                property_type=prop_type,
                options=options
            )
            if prop.value:
                attachment_name = prop._json_data['value'].split('/')[-1]
                with temp_chdir() as target_dir:
                    full_path = os.path.join(target_dir or os.getcwd(), attachment_name)
                    prop.save_as(filename=full_path)
                    moved_prop.upload(full_path)

        # Other properties are quite straightforward
        else:
            moved_prop = moved_part_model.add_property(
                name=prop.name,
                description=description,
                property_type=prop_type,
                default_value=prop.value,
                unit=unit,
                options=options
            )

        # Map the current property model id with newly created property model Object
        get_mapping_dictionary()[prop.id] = moved_prop

    # Now copy the sub-tree of the part
    if include_children:
        # Populate the part so multiple children retrieval is not needed
        part.populate_descendants()
        # For each part, recursively run this function
        for sub_part in part._cached_children:
            relocate_model(
                part=sub_part,
                target_parent=moved_part_model,
                name=sub_part.name,
                include_children=include_children
            )
    return moved_part_model


def get_illegal_targets(part, include):
    """
    Retrieve the illegal parent parts where `Part` can be moved/copied.

    :param part: `Part` to be moved/copied.
    :type part: :class:`Part`
    :param include: `Set` object with id's to be avoided as target parent `Part`
    :type include: set
    :return: `List` object of illegal id's
    :rtype: list
    """
    list_of_illegal_targets = include or set()
    for descendant in part.children(descendants='children'):
        list_of_illegal_targets.add(descendant.id)
    return list_of_illegal_targets


def relocate_instance(part, target_parent, name=None, include_children=True):
    """
    Move the `Part` instance to target parent.

    .. versionadded:: 2.3

    :param part: `Part` object to be moved
    :type part: :class:`Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: `Part` instance
    """
    # First, if the user doesn't provide the name, then just use the default "Clone - ..." name
    if not name:
        name = "CLONE - {}".format(part.name)
    # Initially the model of the part needs to be recreated under the model of the target_parent. Retrieve them.
    part_model = part.model()
    target_parent_model = target_parent.model()

    # Call the move_part() function for those models.
    relocate_model(part=part_model, target_parent=target_parent_model, name=part_model.name,
                   include_children=include_children)

    # Populate the descendants of the Part (category=Instance), in order to avoid to retrieve children for every
    # level and save time. Only need it the children should be included.
    if include_children:
        part.populate_descendants()

    # This function will move the part instance under the target_parent instance, and its children if required.
    moved_instance = move_part_instance(part_instance=part, target_parent=target_parent, part_model=part_model,
                                        name=name, include_children=include_children)
    return moved_instance


def move_part_instance(part_instance, target_parent, part_model, name=None, include_children=True):
    """
    Move the `Part` instance to target parent and updates the properties based on the original part instance.

    .. versionadded:: 2.3

    :param part_instance: `Part` object to be moved
    :type part_instance: :class:`Part`
    :param part_model: `Part` object representing the model of part_instance
    :type part_model: :class: `Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: `Part` instance
    """
    # If no specific name has been required, then call in as Clone of the part_instance.
    if not name:
        name = part_instance.name

    # Retrieve the model of the future part to be created
    moved_model = get_mapping_dictionary()[part_model.id]

    # Now act based on multiplicity
    if moved_model.multiplicity == Multiplicity.ONE:
        # If multiplicity is 'Exactly 1', that means the instance was automatically created with the model.
        moved_instance = moved_model.instances(parent_id=target_parent.id)[0]
    elif moved_model.multiplicity == Multiplicity.ONE_MANY:
        # If multiplicity is '1 or more', that means one instance has automatically been created with the model.
        # Store the model in a list, in case there are multiple instance that need to be recreated.
        if moved_model.id not in get_edited_one_many():
            moved_instance = moved_model.instances(parent_id=target_parent.id)[0]
            get_edited_one_many().append(moved_model.id)
        else:
            moved_instance = target_parent.add(name=name, model=moved_model, suppress_kevents=True)
    else:
        # If multiplicity is '0 or more' or '0 or 1', it means no instance has been created automatically with the
        # model, so then everything must be created and then updated.
        moved_instance = target_parent.add(name=name, model=moved_model, suppress_kevents=True)

    # Update properties of the instance
    map_property_instances(part_instance, moved_instance)
    moved_instance = update_part_with_properties(part_instance, moved_instance, name=str(name))

    # If include_children is True, then recursively call this function for every descendant. Keep the name of the
    # original sub-instance.

    if include_children:
        if part_instance._cached_children is None:
            part_instance.populate_descendants()

        for sub_instance in part_instance._cached_children:
            move_part_instance(part_instance=sub_instance, target_parent=moved_instance,
                               part_model=sub_instance.model(),
                               name=sub_instance.name, include_children=True)

    return moved_instance


def update_part_with_properties(part_instance, moved_instance, name=None):
    """
    Update the newly created part and its properties based on the original one.

    :param part_instance: `Part` object to be copied
    :type part_instance: :class:`Part`
    :param moved_instance: `Part` object copied
    :type moved_instance: :class:`Part`
    :param name: Name of the updated part
    :type name: basestring
    :return: moved :class: `Part` instance
    """
    # Instantiate and empty dictionary later used to map {property.id: property.value} in order to update the part
    # in one go
    properties_id_dict = dict()
    for prop_instance in part_instance.properties:
        # Do different magic if there is an attachment property and it has a value
        if prop_instance.type == PropertyType.ATTACHMENT_VALUE:
            moved_prop = get_mapping_dictionary()[prop_instance.id]
            if prop_instance.value:
                attachment_name = prop_instance._json_data['value'].split('/')[-1]
                with temp_chdir() as target_dir:
                    full_path = os.path.join(target_dir or os.getcwd(), attachment_name)
                    prop_instance.save_as(filename=full_path)
                    moved_prop.upload(full_path)
            else:
                moved_prop.clear()
        # For a reference value property, add the id's of the part referenced {property.id: [part1.id, part2.id, ...]},
        # if there is part referenced at all.
        elif prop_instance.type == PropertyType.REFERENCES_VALUE:
            if prop_instance.value:
                moved_prop_instance = get_mapping_dictionary()[prop_instance.id]
                properties_id_dict[moved_prop_instance.id] = [ref_part.id for ref_part in prop_instance.value]
        else:
            moved_prop_instance = get_mapping_dictionary()[prop_instance.id]
            properties_id_dict[moved_prop_instance.id] = prop_instance.value
    # Update the name and property values in one go.
    moved_instance.update(name=str(name), update_dict=properties_id_dict, bulk=True, suppress_kevents=True)
    moved_instance.populate_descendants()
    return moved_instance


def map_property_instances(original_part, new_part):
    """
    Map the id of the original part with the `Part` object of the newly created one.

    Updated the singleton `mapping dictionary` with the new mapping table values.

    :param original_part: `Part` object to be copied/moved
    :type original_part: :class:`Part`
    :param new_part: `Part` object copied/moved
    :type new_part: :class:`Part`
    :return: None
    """
    # Map the original part with the new one
    get_mapping_dictionary()[original_part.id] = new_part

    # Do the same for each Property of original part instance, using the 'model' id and the get_mapping_dictionary
    for prop_original in original_part.properties:
        get_mapping_dictionary()[prop_original.id] = [
            prop_new for prop_new in new_part.properties if
            get_mapping_dictionary()[prop_original.model_id].id == prop_new.model_id][0]



import datetime
import warnings

from pykechain.exceptions import IllegalArgumentError
from pykechain.models.property import Property
from pykechain.utils import parse_datetime


class DatetimeProperty(Property):
    """A virtual object representing a KE-chain reference property."""

    def __index__(self, json, **kwargs):
        super(DatetimeProperty, self).__init__(json, **kwargs)

    @property
    def value(self):
        """Retrieve the data value of a property.

        Setting this value will immediately update the property in KE-chain.

        :returns: the value
        """
        return self._value

    @value.setter
    def value(self, value):
        if isinstance(value, datetime.datetime):
            if not value.tzinfo:
                warnings.warn("The value '{}' is naive and not timezone aware, use pytz.timezone info. "
                              "This date is interpreted as UTC time.".format(value.isoformat(sep=' ')))

            self._put_value(value.isoformat(sep='T'))
        else:
            raise IllegalArgumentError('value should be a datetime.datetime() object')

    def to_datetime(self):
        """Retrieve the data value of a property.

        Setting this value will immediately update the property in KE-chain.

        :returns: the value
        """
        return parse_datetime(self._value)



import requests
from jsonschema import validate
from six import text_type, string_types

from pykechain.enums import ComponentXType, Category, SortTable, PropertyType, NavigationBarAlignment, WidgetNames, \
    ShowColumnTypes
from pykechain.exceptions import APIError, IllegalArgumentError
from pykechain.models import Part2, Activity2
from pykechain.models.activity import Activity
from pykechain.models.part import Part
from pykechain.models.property import Property
from pykechain.models.service import Service
from pykechain.utils import is_uuid

uuid_pattern = r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
uuid_string = {"type": "string", "pattern": uuid_pattern}

# deprecated as we are moving towards a 'meta' definition for the widgets in KE-chain 3
# only for checking the jsonwidget
component_jsonwidget_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "Component JSON Schema",
    "type": "object",
    "properties": {
        "xtype": {
            "type": "string",
            "enum": ComponentXType.values()
        },
        "filter": {
            "type": "object",
            "properties": {
                "part": uuid_string,
                "model": uuid_string,
                "parent": uuid_string,
                "part_id": uuid_string,
                "model_id": uuid_string,
                "parent_id": uuid_string
            }
        },
        "title": {"type": ["string", "null"]},
        "viewModel": {"type": "object"},
        "model": uuid_string,
        "parent": uuid_string
    },
    "required": ["xtype"]
}

widget_json_stub = {
    "title": "Widget JSON schema",
    "type": "object",
    "required": ["name", "meta"],
    "properties": {
        "name": {"type": "string", "enum": WidgetNames.values()},
        "meta": {"type": "object"},
        "config": {"type": "object"}
    }
}

widgetconfig_json_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "WidgetConfig JSON",
    "type": "object",
    "properties": {
        "ext": {
            "type": "object",
            "widgets": {
                "type": "array",
                "items": widget_json_stub
            }
        }
    }
}


class CustomizationBase(object):
    """Base class for customization objects.

    :cvar activity: an instance of :class:`pykechain.models.Activity`
    """

    def __init__(self, activity, client):
        """Initialize the Base class for customization objects."""
        self._client = client
        self.activity = activity


class ExtCustomization(CustomizationBase):
    """A class to represent the activity customization for Ext Js."""

    def __str__(self):  # pragma: no cover
        return "<pyke ExtCustomization '{}' id {} ({} widgets)>".format(
            self.activity.name, str(self.activity.id)[-8:], len(self.widgets()))

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    def _save_customization(self, widgets):
        """
        Save the complete customization to the activity.

        :param widgets: The complete set of widgets to be customized
        """
        if len(widgets) > 0:
            # Get the current customization and only replace the 'ext' part of it
            customization = self.activity._json_data.get('customization', dict())
            if customization:
                customization['ext'] = dict(widgets=widgets)
            else:
                customization = dict(ext=dict(widgets=widgets))

        # Empty the customization if if the widgets list is empty
        else:
            customization = None

        # perform validation
        if customization:
            validate(customization, widgetconfig_json_schema)

        # Save to the activity and store the saved activity to self
        response = self._client._request("PUT",
                                         self._client._build_url("activity", activity_id=str(self.activity.id)),
                                         json=dict(customization=customization))
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not save customization ({})".format(response))
        else:
            # refresh the activity json
            self.activity = self._client.activity(pk=self.activity.id)

    def _add_widget(self, widget):
        """
        Add a widget to the customization.

        Will save the widget to KE-chain.

        :param widget: The widget (specific json dict) to be added
        :type widget: dict
        """
        widgets = self.widgets()
        widgets += [widget]
        self._save_customization(widgets)

    def widgets(self):
        """
        Get the Ext JS specific customization from the activity.

        :return: The Ext JS specific customization in `list(dict)` form
        """
        customization = self.activity._json_data.get('customization')

        if customization and "ext" in customization.keys():
            return customization['ext']['widgets']
        else:
            return []

    def delete_widget(self, index):
        """
        Delete widgets by index.

        The widgets are saved to KE-chain.

        :param index: The index of the widget to be deleted in the self.widgets
        :type index: int
        :raises ValueError: if the customization has no widgets
        """
        widgets = self.widgets()
        if len(widgets) == 0:
            raise ValueError("This customization has no widgets")
        widgets.pop(index)
        self._save_customization(widgets)

    def delete_all_widgets(self):
        """Delete all widgets.

        :raises APIError: if the widgets could not be deleted.
        """
        self._save_customization([])

    def add_json_widget(self, config):
        """
        Add an Ext Json Widget to the customization.

        The configuration json provided must be interpretable by KE-chain. The json will be validated
        against the widget json schema.

        The widget will be saved to KE-chain.

        :param config: The json configuration of the widget
        :type config: dict
        """
        validate(config, component_jsonwidget_schema)
        self._add_widget(dict(config=config, name=WidgetNames.JSONWIDGET))

    def add_super_grid_widget(self, part_model, delete=False, edit=True, export=True, incomplete_rows=True,
                              new_instance=False, parent_part_instance=None, max_height=None, custom_title=False,
                              emphasize_edit=False, emphasize_new_instance=True, sort_property=None,
                              sort_direction=SortTable.ASCENDING):
        """
        Add a KE-chain superGrid (e.g. basic table widget) to the customization.

        The widget will be saved to KE-chain.

        :param emphasize_new_instance: Emphasize the New instance button (default True)
        :type emphasize_new_instance: bool
        :param emphasize_edit: Emphasize the Edit button (default False)
        :type emphasize_edit: bool
        :param new_instance: Show or hide the New instance button (default False). You need to provide a
            `parent_part_instance` in order for this to work.
        :type new_instance: bool
        :param incomplete_rows: Show or hide the Incomplete Rows filter button (default True)
        :type incomplete_rows: bool
        :param export: Show or hide the Export Grid button (default True)
        :type export: bool
        :param edit: Show or hide the Edit button (default True)
        :type edit: bool
        :param delete: Show or hide the Delete button (default False)
        :type delete: bool
        :param part_model: The part model based on which all instances will be shown.
        :type parent_part_instance: :class:`Part` or UUID
        :param parent_part_instance: The parent part instance for which the instances will be shown or to which new
            instances will be added.
        :type parent_part_instance: :class:`Part` or UUID
        :param max_height: The max height of the supergrid in pixels
        :type max_height: int or None
        :param custom_title: A custom title for the supergrid::
            * False (default): Part instance name
            * String value: Custom title
            * None: No title
        :type custom_title: bool or basestring or None
        :param sort_property: The property model on which the part instances are being sorted on
        :type sort_property: :class:`Property` or UUID
        :param sort_direction: The direction on which the values of property instances are being sorted on:
            * ASC (default): Sort in ascending order
            * DESC: Sort in descending order
        :type sort_direction: basestring (see :class:`enums.SortTable`)
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.

        """
        # Check whether the part_model is uuid type or class `Part`
        if isinstance(part_model, (Part, Part2)):
            part_model_id = part_model.id
        elif isinstance(part_model, (string_types, text_type)) and is_uuid(part_model):
            part_model_id = part_model
            part_model = self._client.model(id=part_model_id)
        else:
            raise IllegalArgumentError("When using the add_supergrid_widget, part_model must be a Part or Part id. "
                                       "Type is: {}".format(type(part_model)))

        # Check whether the parent_part_instance is uuid type or class `Part`
        if isinstance(parent_part_instance, (Part, Part2)):
            parent_part_instance_id = parent_part_instance.id
        elif isinstance(parent_part_instance, (string_types, text_type)) and is_uuid(parent_part_instance):
            parent_part_instance_id = parent_part_instance
            parent_part_instance = self._client.part(id=parent_part_instance_id)
        elif isinstance(parent_part_instance, type(None)):
            parent_part_instance_id = None
        else:
            raise IllegalArgumentError("When using the add_supergrid_widget, parent_part_instance must be a "
                                       "Part, Part id or None. Type is: {}".format(type(parent_part_instance)))

        # Check whether the sort_property is uuid type or class `Property`
        if isinstance(sort_property, Property):
            sort_property_id = sort_property.id
        elif isinstance(sort_property, (string_types, text_type)) and is_uuid(sort_property):
            sort_property_id = sort_property
            sort_property = self._client.property(id=sort_property_id, category=Category.MODEL)
        elif isinstance(sort_property, type(None)):
            sort_property_id = None
        else:
            raise IllegalArgumentError("When using the add_supergrid_widget, sort_property must be a "
                                       "Property, Property id or None. Type is: {}".format(type(sort_property)))

        # Assertions
        if not parent_part_instance and new_instance:
            raise IllegalArgumentError("If you want to allow the creation of new part instances, you must specify a "
                                       "parent_part_instance")
        if sort_property and sort_property.part.id != part_model.id:
            raise IllegalArgumentError("If you want to sort on a property, then sort_property must be located under "
                                       "part_model")
        # Declare superGrid config
        if self._client.match_app_version(label='pim', version='>=3.0.0'):
            config = {
                "xtype": ComponentXType.SUPERGRID,
                "filter": {
                    "activity_id": str(self.activity.id),
                    "model_id": part_model_id
                },
                "viewModel": {
                    "data": {
                        "actions": {
                            "delete": delete,
                            "edit": edit,
                            "export": export,
                            "incompleteRows": incomplete_rows,
                            "newInstance": new_instance
                        },
                        "sorters": [{
                            "direction": sort_direction,
                            "property": sort_property_id
                        }] if sort_property_id else [],
                        "ui": {
                            "newInstance": "primary-action" if emphasize_new_instance else "default-toolbar",
                            "edit": "primary-action" if emphasize_edit else "default-toolbar"
                        }
                    }
                }
            }
            # Add parent to filter if added.
            config['filter']["parent_id"] = parent_part_instance_id
        else:
            config = {
                "xtype": ComponentXType.SUPERGRID,
                "filter": {
                    "activity_id": str(self.activity.id),
                    "model": part_model_id
                },
                "viewModel": {
                    "data": {
                        "actions": {
                            "delete": delete,
                            "edit": edit,
                            "export": export,
                            "incompleteRows": incomplete_rows,
                            "newInstance": new_instance
                        },
                        "sorters": [{
                            "direction": sort_direction,
                            "property": sort_property_id
                        }] if sort_property_id else [],
                        "ui": {
                            "newInstance": "primary-action" if emphasize_new_instance else "default-toolbar",
                            "edit": "primary-action" if emphasize_edit else "default-toolbar"
                        }
                    }
                }
            }
            config['filter']["parent"] = parent_part_instance_id

        # Add max height and custom title
        if max_height:
            config['maxHeight'] = max_height

        if custom_title is False:
            show_title_value = "Default"
            title = part_model.name
        elif custom_title is None:
            show_title_value = "No title"
            title = str()
        else:
            show_title_value = "Custom Title"
            title = str(custom_title)

        config["title"] = title
        config["showTitleValue"] = show_title_value

        # Declare the meta info for the superGrid
        meta = {
            "parentInstanceId": parent_part_instance_id,
            "editButtonUi": "primary-action" if emphasize_edit else "default-toolbar",
            "customHeight": max_height if max_height else 500,
            "primaryAddUiValue": emphasize_new_instance,
            "activityId": str(self.activity.id),
            "customTitle": title,
            "primaryEditUiValue": emphasize_edit,
            "downloadButtonVisible": export,
            "addButtonUi": "primary-action" if emphasize_new_instance else "default-toolbar",
            "deleteButtonVisible": delete,
            "incompleteRowsButtonVisible": incomplete_rows,
            "addButtonVisible": new_instance,
            "showTitleValue": show_title_value,
            "partModelId": part_model_id,
            "editButtonVisible": edit,
            "showHeightValue": "Custom max height" if max_height else "Auto",
            "sortDirection": sort_direction,
            "sortedColumn": sort_property_id if sort_property_id else None
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.SUPERGRIDWIDGET))

    def add_property_grid_widget(self, part_instance, max_height=None, custom_title=False, show_headers=True,
                                 show_columns=None):
        """
        Add a KE-chain Property Grid widget to the customization.

        The widget will be saved to KE-chain.

        :param part_instance: The part instance on which the property grid will be based
        :type part_instance: :class:`Part` or UUID
        :param max_height: The max height of the property grid in pixels
        :type max_height: int or None
        :param custom_title: A custom title for the property grid::
            * False (default): Part instance name
            * String value: Custom title
            * None: No title
        :type custom_title: bool or basestring or None
        :param show_headers: Show or hide the headers in the grid (default True)
        :type show_headers: bool
        :param show_columns: Columns to be hidden or shown (default to 'unit' and 'description')
        :type show_columns: list
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Check whether the parent_part_instance is uuid type or class `Part`
        if isinstance(part_instance, (Part, Part2)):
            part_instance_id = part_instance.id
        elif isinstance(part_instance, (string_types, text_type)) and is_uuid(part_instance):
            part_instance_id = part_instance
            part_instance = self._client.part(id=part_instance_id)
        else:
            raise IllegalArgumentError("When using the add_property_grid_widget, part_instance must be a "
                                       "Part or Part id. Type is: {}".format(type(part_instance)))
        if not show_columns:
            show_columns = list()

        # Set the display_columns for the config
        possible_columns = [ShowColumnTypes.DESCRIPTION, ShowColumnTypes.UNIT]
        display_columns = dict()

        for possible_column in possible_columns:
            if possible_column in show_columns:
                display_columns[possible_column] = True
            else:
                display_columns[possible_column] = False

        # Declare property grid config
        if self._client.match_app_version(label='pim', version='>=3.0.0'):
            config = {
                "xtype": ComponentXType.PROPERTYGRID,
                "category": Category.INSTANCE,
                "filter": {
                    "activity_id": str(self.activity.id),
                    "part_id": part_instance_id  # gpim
                },
                "hideHeaders": not show_headers,
                "viewModel": {
                    "data": {
                        "displayColumns": display_columns
                    }
                }
            }
        else:
            config = {
                "xtype": ComponentXType.PROPERTYGRID,
                "category": Category.INSTANCE,
                "filter": {
                    "activity_id": str(self.activity.id),
                    "part": part_instance_id
                },
                "hideHeaders": not show_headers,
                "viewModel": {
                    "data": {
                        "displayColumns": display_columns
                    }
                }
            }

        # Add max height and custom title
        if max_height:
            config['maxHeight'] = max_height

        if custom_title is False:
            show_title_value = "Default"
            title = part_instance.name
        elif custom_title is None:
            show_title_value = "No title"
            title = str()
        else:
            show_title_value = "Custom Title"
            title = str(custom_title)

        config["title"] = title

        # Declare the meta info for the property grid
        meta = {
            "activityId": str(self.activity.id),
            "customHeight": max_height if max_height else None,
            "customTitle": title,
            "partInstanceId": part_instance_id,
            "showColumns": show_columns,
            "showHeaders": show_headers,
            "showHeightValue": "Custom max height" if max_height else "Auto",
            "showTitleValue": show_title_value
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.PROPERTYGRIDWIDGET))

    def add_text_widget(self, text=None, custom_title=None, collapsible=True, collapsed=False):
        """
        Add a KE-chain Text widget to the customization.

        The widget will be saved to KE-chain.

        :param text: The text that will be shown by the widget.
        :type text: basestring or None
        :param custom_title: A custom title for the text panel::
            * None (default): No title
            * String value: Custom title
        :type custom_title: basestring or None
        :param collapsible: A boolean to decide whether the panel is collapsible or not (default True)
        :type collapsible: bool
        :param collapsed: A boolean to decide whether the panel is collapsed or not (default False)
        :type collapsible: bool
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Declare text widget config
        config = {
            "xtype": ComponentXType.HTMLPANEL,
            "filter": {
                "activity_id": str(self.activity.id),
            }
        }

        # Add text and custom title
        if text:
            config['html'] = text
        if custom_title:
            show_title_value = "Custom Title"
            title = custom_title
        else:
            show_title_value = "No Title"
            title = None
        config['collapsible'] = collapsible
        # A widget can only be collapsed if it is collapsible in the first place
        if collapsible:
            config['collapsed'] = collapsed
        else:
            config['collapsed'] = False
        config['title'] = title
        # Declare the meta info for the property grid
        meta = {
            "activityId": str(self.activity.id),
            "customTitle": title,
            "collapsible": collapsible,
            "collapsed": collapsed,
            "html": text,
            "showTitleValue": show_title_value
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.HTMLWIDGET))

    def add_paginated_grid_widget(self, part_model, delete=False, edit=True, export=True,
                                  new_instance=False, clone_instance=False, parent_part_instance=None, max_height=None,
                                  custom_title=False, emphasize_edit=False, emphasize_new_instance=True,
                                  emphasize_clone_instance=False, sort_property=None,
                                  sort_direction=SortTable.ASCENDING, page_size=25, collapse_filters=False):
        """
        Add a KE-chain paginatedGrid (e.g. paginated table widget) to the customization.

        The widget will be saved to KE-chain.

        :param new_instance: Show or hide the New instance button (default False). You need to provide a
            `parent_part_instance` in order for this to work.
        :type new_instance: bool
        :param clone_instance: Show or hide the Clone instance button (default False). You need to provide a
            `parent_part_instance` in order for this to work.
        :type clone_instance: bool
        :param edit: Show or hide the Edit button (default True)
        :type edit: bool
        :param delete: Show or hide the Delete button (default False)
        :type delete: bool

        :param collapse_filters: Hide or show the filters pane (default False)
        :type collapse_filters: bool
        :param incomplete_rows: Show or hide the Incomplete Rows filter button (default True)
        :type incomplete_rows: bool
        :param export: Show or hide the Export Grid button (default True)
        :type export: bool

        :param part_model: The part model based on which all instances will be shown.
        :param part_model: :class:`Part` or UUID
        :param parent_part_instance: The parent part instance for which the instances will be shown or to which new
            instances will be added.
        :type parent_part_instance: :class:`Part` or UUID

        :param emphasize_new_instance: Emphasize the New instance button (default True)
        :type emphasize_new_instance: bool
        :param emphasize_edit: Emphasize the Edit button (default False)
        :type emphasize_edit: bool
        :param emphasize_clone_instance:  emphasize the Clone instance button (default False)
        :type emphasize_clone_instance: bool

        :param page_size: Number of parts that will be shown per page in the grid.
        :type page_size: int
        :param max_height: The max height of the paginated grid in pixels
        :type max_height: int or None
        :param custom_title: A custom title for the paginated grid::
            * False (default): Part instance name
            * String value: Custom title
            * None: No title
        :type custom_title: bool or basestring or None
        :param sort_property: The property model on which the part instances are being sorted on
        :type sort_property: :class:`Property` or UUID
        :param sort_direction: The direction on which the values of property instances are being sorted on::
            * ASC (default): Sort in ascending order
            * DESC: Sort in descending order
        :type sort_direction: basestring (see :class:`enums.SortTable`)
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Check whether the part_model is uuid type or class `Part`
        if isinstance(part_model, (Part, Part2)):
            part_model_id = part_model.id
        elif isinstance(part_model, (string_types, text_type)) and is_uuid(part_model):
            part_model_id = part_model
            part_model = self._client.model(id=part_model_id)
        else:
            raise IllegalArgumentError("When using the add_paginated_grid_widget, part_model must be a Part or Part id."
                                       " Type is: {}".format(type(part_model)))

        # Check whether the parent_part_instance is uuid type or class `Part`
        if isinstance(parent_part_instance, (Part, Part2)):
            parent_part_instance_id = parent_part_instance.id
        elif isinstance(parent_part_instance, (string_types, text_type)) and is_uuid(parent_part_instance):
            parent_part_instance_id = parent_part_instance
            parent_part_instance = self._client.part(id=parent_part_instance_id)
        elif isinstance(parent_part_instance, type(None)):
            parent_part_instance_id = None
        else:
            raise IllegalArgumentError("When using the add_paginated_grid_widget, parent_part_instance must be a "
                                       "Part, Part id or None. Type is: {}".format(type(parent_part_instance)))

        # Check whether the sort_property is uuid type or class `Property`
        if isinstance(sort_property, Property):
            sort_property_id = sort_property.id
        elif isinstance(sort_property, (string_types, text_type)) and is_uuid(sort_property):
            sort_property_id = sort_property
            sort_property = self._client.property(id=sort_property_id, category=Category.MODEL)
        elif isinstance(sort_property, type(None)):
            sort_property_id = None
        else:
            raise IllegalArgumentError("When using the add_paginated_grid_widget, sort_property must be a "
                                       "Property, Property id or None. Type is: {}".format(type(sort_property)))

        # Assertions
        if not parent_part_instance and (new_instance or clone_instance):
            raise IllegalArgumentError("If you want to allow the creation (or cloning) of new part instances, "
                                       "you must specify a `parent_part_instance`")
        if sort_property and sort_property.part.id != part_model.id:
            raise IllegalArgumentError("If you want to sort on a property, then sort_property must be located under "
                                       "`part_model`")

        # Add custom title
        if custom_title is False:
            show_title_value = "Default"
            title = part_model.name
        elif custom_title is None:
            show_title_value = "No title"
            title = ' '
        else:
            show_title_value = "Custom Title"
            title = str(custom_title)

        # Set the collapse filters value
        if collapse_filters:
            collapse_filters_value = "Collapsed"
        else:
            collapse_filters_value = "Expanded"

        config = {
            "xtype": ComponentXType.FILTEREDGRID,
            "filter": {
                "activity_id": str(self.activity.id),
            },
            "grid": {
                "viewModel": {
                    "data": {
                        "actions": {
                            "delete": delete,
                            "edit": edit,
                            "export": export,
                            "newInstance": new_instance,
                            "cloneInstance": clone_instance
                        },
                        "sorters": [{
                            "direction": sort_direction,
                            "property": sort_property_id
                        }] if sort_property_id else [],
                        "ui": {
                            "newInstance": "primary-action" if emphasize_new_instance else "default-toolbar",
                            "edit": "primary-action" if emphasize_edit else "default-toolbar",
                            "cloneInstance": "primary-action" if emphasize_clone_instance else "default-toolbar"
                        },
                        "pageSize": page_size
                    }
                },
                "xtype": ComponentXType.PAGINATEDSUPERGRID,
                "title": title,
                "showTitleValue": show_title_value,
            },
            "maxHeight": max_height if max_height else None,
            "parentInstanceId": parent_part_instance_id,
            "partModelId": part_model_id,
            "collapseFilters": collapse_filters
        }

        # Declare the meta info for the paginatedGrid
        meta = {
            "parentInstanceId": parent_part_instance_id,
            "editButtonUi": "primary-action" if emphasize_edit else "default-toolbar",
            "editButtonVisible": edit,
            "customHeight": max_height if max_height else 500,
            "primaryAddUiValue": emphasize_new_instance,
            "activityId": str(self.activity.id),
            "customTitle": title,
            "primaryEditUiValue": emphasize_edit,
            "downloadButtonVisible": export,
            "addButtonUi": "primary-action" if emphasize_new_instance else "default-toolbar",
            "deleteButtonVisible": delete,
            "addButtonVisible": new_instance,
            "showTitleValue": show_title_value,
            "partModelId": str(part_model_id),
            "showHeightValue": "Custom max height" if max_height else "Auto",
            "sortDirection": sort_direction,
            "sortedColumn": sort_property_id if sort_property_id else None,
            "collapseFilters": collapse_filters,
            "showCollapseFiltersValue": collapse_filters_value,
            "customPageSize": page_size
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.FILTEREDGRIDWIDGET))

    def add_script_widget(self, script, custom_title=False, custom_button_text=False, emphasize_run=True):
        """
        Add a KE-chain Script (e.g. script widget) to the customization.

        The widget will be saved to KE-chain.

        :param script: The Script to which the button will be coupled and will be ran when the button is pressed.
        :type script: :class:`Service` or UUID
        :param custom_title: A custom title for the script widget
            * False (default): Script name
            * String value: Custom title
            * None: No title
        :type custom_title: bool or basestring or None
        :param custom_button_text: A custom text for the button linked to the script
            * False (default): Script name
            * String value: Custom title
            * None: No title
        :type custom_button_text: bool or basestring or None
        :param emphasize_run: Emphasize the run button (default True)
        :type emphasize_run: bool
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Check whether the script is uuid type or class `Service`
        if isinstance(script, Service):
            script_id = script.id
        elif isinstance(script, (string_types, text_type)) and is_uuid(script):
            script_id = script
            script = self._client.service(id=script_id)
        else:
            raise IllegalArgumentError("When using the add_script_widget, script must be a Service or Service id. "
                                       "Type is: {}".format(type(script)))
        # Add custom title
        if custom_title is False:
            show_title_value = "Default"
            title = script.name
        elif custom_title is None:
            show_title_value = "No title"
            title = ''
        else:
            show_title_value = "Custom Title"
            title = str(custom_title)

        # Add custom button text
        if custom_button_text is False:
            show_button_value = "Default"
            button_text = script.name
        elif custom_button_text is None:
            show_button_value = "No title"
            button_text = ''
        else:
            show_button_value = "Custom Title"
            button_text = str(custom_button_text)

        # Declare script widget config
        config = {
            'showTitleValue': show_title_value,
            'serviceId': script_id,
            'viewModel': {
                'data': {
                    'buttonUI': 'primary-action' if emphasize_run else "default-toolbar",
                }
            },
            'title': title,
            'xtype': ComponentXType.EXECUTESERVICE,
            'customButtonText': button_text
        }

        # Declare script widget meta
        meta = {
            'showButtonValue': show_button_value,
            'customText': button_text,
            'customTitle': title,
            'serviceId': script_id,
            'emphasizeButton': emphasize_run,
            'showTitleValue': show_title_value
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.SERVICEWIDGET))

    def add_notebook_widget(self, notebook, custom_title=False, height=None):
        """
        Add a KE-chain Notebook (e.g. notebook widget) to the customization.

        The widget will be saved to KE-chain.

        :param notebook: The Notebook to which the button will be coupled and will start when the button is pressed.
        :type notebook: :class:`Service` or UUID
        :param custom_title: A custom title for the notebook widget
            * False (default): Notebook name
            * String value: Custom title
            * None: No title
        :type custom_title: bool or basestring or None
        :param height: The height of the Notebook in pixels
        :type height: int or None
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Check whether the script is uuid type or class `Service`
        if isinstance(notebook, Service):
            notebook_id = notebook.id
        elif isinstance(notebook, (string_types, text_type)) and is_uuid(notebook):
            notebook_id = notebook
            notebook = self._client.service(id=notebook_id)
        else:
            raise IllegalArgumentError("When using the add_notebook_widget, notebook must be a Service or Service id. "
                                       "Type is: {}".format(type(notebook)))

        # Add custom title
        if custom_title is False:
            show_title_value = "Default"
            title = notebook.name
        elif custom_title is None:
            show_title_value = "No title"
            title = ''
        else:
            show_title_value = "Custom Title"
            title = str(custom_title)

        # Declare notebook widget config
        config = {
            'title': title,
            'showTitleValue': show_title_value,
            'height': height if height else 800,
            'xtype': ComponentXType.NOTEBOOKPANEL,
            'serviceId': notebook_id
        }

        # Declare notebook widget meta
        meta = {
            'showTitleValue': show_title_value,
            'customHeight': height if height else 800,
            'customTitle': title,
            'serviceId': notebook_id
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.NOTEBOOKWIDGET))

    def add_attachment_viewer_widget(self, attachment_property, custom_title=False, height=None):
        """
        Add a KE-chain Attachment Viewer (e.g. attachment viewer widget) to the customization.

        The widget will be saved to KE-chain.

        :param attachment_property: The Attachment Property to which the Viewer will be connected to.
        :type attachment_property: :class:`Property` or UUID
        :param custom_title: A custom title for the attachment viewer widget
            * False (default): Notebook name
            * String value: Custom title
            * None: No title
        :type custom_title: bool or basestring or None
        :param height: The height of the Notebook in pixels
        :type height: int or None
        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Check whether the attachment property is uuid type or class `Property`
        if isinstance(attachment_property, Property):
            attachment_property_id = attachment_property.id
        elif isinstance(attachment_property, (string_types, text_type)) and is_uuid(attachment_property):
            attachment_property_id = attachment_property
            attachment_property = self._client.property(id=attachment_property_id)
        else:
            raise IllegalArgumentError("When using the add_attachment_viewer_widget, attachment_property must be a "
                                       "Property or Property id. Type is: {}".format(type(attachment_property)))

        # Check whether the `Property` has type `Attachment`
        property_type = attachment_property.type
        if property_type != PropertyType.ATTACHMENT_VALUE:
            raise IllegalArgumentError("When using the add_attachment_viewer_widget, attachment_property must have "
                                       "type {}. Type found: {}".format(PropertyType.ATTACHMENT_VALUE, property_type))

        # Check also whether `Property` has category `Instance`
        property_category = attachment_property._json_data['category']
        if property_category != Category.INSTANCE:
            raise IllegalArgumentError("When using the add_attachment_viewer_widget, attachment_property must have "
                                       "category {}. Category found: {}".format(Category.INSTANCE, property_category))

        # Add custom title
        if custom_title is False:
            show_title_value = "Default"
            title = attachment_property.name
        elif custom_title is None:
            show_title_value = "No title"
            title = ''
        else:
            show_title_value = "Custom Title"
            title = str(custom_title)

        # Declare attachment viewer widget config
        config = {
            'propertyId': attachment_property_id,
            'showTitleValue': show_title_value,
            'xtype': ComponentXType.PROPERTYATTACHMENTPREVIEWER,
            'title': title,
            'filter': {
                'activity_id': str(self.activity.id)
            },
            'height': height if height else 500
        }

        # Declare attachment viewer widget meta
        meta = {
            'propertyInstanceId': attachment_property_id,
            'activityId': str(self.activity.id),
            'customHeight': height if height else 500,
            'showTitleValue': show_title_value,
            'customTitle': title
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.ATTACHMENTVIEWERWIDGET))

    def add_navigation_bar_widget(self, activities, alignment=NavigationBarAlignment.CENTER):
        """
        Add a KE-chain Navigation Bar (e.g. navigation bar widget) to the customization.

        The widget will be saved to KE-chain.

        :param activities: List of activities. Each activity must be a Python dict(), with the following keys:
            * customText: A custom text for each button in the attachment viewer widget: None (default): Task name;
            a String value: Custom text
            * emphasize: bool which determines if the button should stand-out or not - default(False)
            * activityId: class `Activity` or UUID
        :type activities: list
        :param alignment: The alignment of the buttons inside navigation bar
            * center (default): Center aligned
            * start: left aligned
        :type alignment: basestring (see :class:`enums.NavigationBarAlignment`)

        :raises IllegalArgumentError: When unknown or illegal arguments are passed.
        """
        # Loop through the list of activities
        set_of_expected_keys = {'activityId', 'customText', 'emphasize'}
        for activity_dict in activities:
            if set(activity_dict.keys()).issubset(set_of_expected_keys) and 'activityId' in set_of_expected_keys:
                # Check whether the activityId is class `Activity` or UUID
                activity = activity_dict['activityId']
                if isinstance(activity, (Activity, Activity2)):
                    activity_dict['activityId'] = activity.id
                elif isinstance(activity, (string_types, text_type)) and is_uuid(activity):
                    pass
                else:
                    raise IllegalArgumentError("When using the add_navigation_bar_widget, activityId must be an "
                                               "Activity or Activity id. Type is: {}".format(type(activity)))
                if 'customText' not in activity_dict.keys() or not activity_dict['customText']:
                    activity_dict['customText'] = str()
                if 'emphasize' not in activity_dict.keys():
                    activity_dict['emphasize'] = False

            else:
                raise IllegalArgumentError("Found unexpected key in activities. Only keys allowed are: {}".
                                           format(set_of_expected_keys))

        # Declare navigation bar widget config
        config = {
            'alignment': alignment,
            'xtype': ComponentXType.ACTIVITYNAVIGATIONBAR,
            'filter': {
                'activity_id': str(self.activity.id)
            },
            'taskButtons': activities
        }

        for activity_dict in activities:
            activity_id = activity_dict['activityId']
            activity_dict['name'] = self._client.activity(id=activity_id).name

        # Declare navigation bar widget meta
        meta = {
            'alignment': alignment,
            'activityId': str(self.activity.id),
            'taskButtons': activities
        }

        self._add_widget(dict(config=config, meta=meta, name=WidgetNames.TASKNAVIGATIONBARWIDGET))



from requests import Response


class APIError(Exception):
    """A general KE-chain API Error occurred.

    A end-user descriptive message is required.

    :ivar response: response object
    :ivar request: request object that preceedes the response
    :ivar msg: error message in the response
    :ivar traceback: traceback in the response (from the KE-chain server)
    :ivar detail: details of the error
    """

    def __init__(self, *args, **kwargs):
        """Initialise the `APIError` with `response`, `request`, `msg`, `traceback` and `detail`.

        :param response:
        :param kwargs:
        """
        response = kwargs.pop('response', None)
        self.response = response
        self.request = kwargs.pop('request', None)
        if (response is not None and not self.request and hasattr(response, 'request')):
            self.request = self.response.request
        self.msg, self.traceback, self.detail = None, None, None
        super(APIError, self).__init__(*args, **kwargs)

        if isinstance(self.response, Response) and 'content' in self.response:
            # may be a KE-chain API response error
            if self.response.json():
                self.msg = self.response.json().get('msg')
                self.traceback = self.response.json().get('traceback')
                self.detail = self.response.json().get('detail')
            else:
                self.traceback = self.response.text
                self.msg = self.traceback.split(r'\n'[-1])


class ForbiddenError(APIError):
    """A login is required."""

    pass


class MultipleFoundError(APIError):
    """Multiple objects are found, while a single object is requested."""

    pass


class NotFoundError(APIError):
    """No object is found."""

    pass


class ClientError(APIError):
    """When instantiating the Client an Error occurred."""

    pass


class IllegalArgumentError(ValueError):
    """Illegal arguments where provided."""

    pass


class InspectorComponentError(Exception):
    """Error in the InspectorComponent."""

    pass



import json

import requests
from six import text_type, string_types
from typing import Any, AnyStr  # noqa: F401

from pykechain.enums import Multiplicity, Category
from pykechain.exceptions import NotFoundError, APIError, MultipleFoundError, IllegalArgumentError
from pykechain.extra_utils import relocate_model, relocate_instance, move_part_instance
from pykechain.models.base import Base
from pykechain.models.property import Property
from pykechain.utils import find, is_uuid


class Part(Base):
    """A virtual object representing a KE-chain part.

    :cvar basestring category: The category of the part, either 'MODEL' or 'INSTANCE'
                               (use :class:`pykechain.enums.Category`)
    :cvar basestring parent_id: The UUID of the parent of this part
    :cvar properties: The list of :class:`Property` objects belonging to this part.
    :cvar basestring multiplicity: The multiplicity of the part being one of the following options:
                                   ZERO_ONE, ONE, ZERO_MANY, ONE_MANY, (reserved) M_N

    Examples
    --------
    For the category property

    >>> bike = project.part('Bike')
    >>> bike.category
    'INSTANCE'

    >>> bike_model = project.model('Bike')
    >>> bike_model.category
    'MODEL'

    >>> bike_model == Category.MODEL
    True
    >>> bike == Category.INSTANCE
    True

    For the multiplicity property

    >>> bike = project.models('Bike')
    >>> bike.multiplicity
    ONE_MANY

    >>> from pykechain.enums import Multiplicity
    >>> bike.multiplicity == Multiplicity.ONE_MANY
    True

    """

    def __init__(self, json, **kwargs):
        # type: (dict, **Any) -> None
        """Construct a part from a KE-chain 2 json response.

        :param json: the json response to construct the :class:`Part` from
        :type json: dict
        """
        super(Part, self).__init__(json, **kwargs)

        self.category = json.get('category')
        self.parent_id = json['parent'].get('id') if 'parent' in json and json.get('parent') else None
        self.properties = [Property.create(p, client=self._client) for p in json['properties']]
        self.multiplicity = json.get('multiplicity', None)
        self._cached_children = None

    def property(self, name):
        """Retrieve the property belonging to this part based on its name or uuid.

        :param name: property name or property UUID to search for
        :type name: basestring
        :return: a single :class:`Property`
        :raises NotFoundError: if the `Property` is not part of the `Part`

        Example
        -------
        >>> part = project.part('Bike')
        >>> part.properties
        [<pyke Property ...>, ...]
        # this returns a list of all properties of this part

        >>> gears = part.property('Gears')
        >>> gears.value
        6

        >>> gears = part.property('123e4567-e89b-12d3-a456-426655440000')
        >>> gears.value
        6

        """
        found = None
        if is_uuid(name):
            found = find(self.properties, lambda p: name == p.id)
        else:
            found = find(self.properties, lambda p: name == p.name)

        if not found:
            raise NotFoundError("Could not find property with name or id {}".format(name))

        return found

    def parent(self):
        # type: () -> Any
        """Retrieve the parent of this `Part`.

        :return: the parent :class:`Part` of this part
        :raises APIError: if an Error occurs

        Example
        -------
        >>> part = project.part('Frame')
        >>> bike = part.parent()

        """
        if self.parent_id:
            return self._client.part(pk=self.parent_id, category=self.category)
        else:
            return None

    def children(self, **kwargs):
        """Retrieve the children of this `Part` as `Partset`.

        When you call the :func:`Part.children()` method without any additional filtering options for the children,
        the children are cached to help speed up subsequent calls to retrieve the children. The cached children are
        returned as a list and not as a `Partset`.

        When you *do provide* additional keyword arguments (kwargs) that act as a specific children filter, the
        cached children are _not_ used and a separate API call is made to retrieve only those children.

        :param kwargs: Additional search arguments to search for, check :class:`pykechain.Client.parts`
                       for additional info
        :type kwargs: dict
        :return: a set of `Parts` as a :class:`PartSet`. Will be empty if no children. Will be a `List` if the
                 children are retrieved from the cached children.
        :raises APIError: When an error occurs.

        Example
        -------
        A normal call, which caches all children of the bike. If you call `bike.children` twice only 1 API call is made.

        >>> bike = project.part('Bike')
        >>> direct_descendants_of_bike = bike.children()

        An example with providing additional part search parameters 'name__icontains'. Children are retrieved from the
        API, not the bike's internal (already cached in previous example) cache.

        >>> bike = project.part('Bike')
        >>> wheel_children_of_bike = bike.children(name__icontains='wheel')

        """
        if not kwargs:
            # no kwargs provided is the default, we aim to cache it.
            if not self._cached_children:
                self._cached_children = list(self._client.parts(parent=self.id, category=self.category))
            return self._cached_children
        else:
            # if kwargs are provided, we assume no use of cache as specific filtering on the children is performed.
            return self._client.parts(parent=self.id, category=self.category, **kwargs)

    def siblings(self, **kwargs):
        # type: (Any) -> Any
        """Retrieve the siblings of this `Part` as `Partset`.

        Siblings are other Parts sharing the same parent of this `Part`, including the part itself.

        :param kwargs: Additional search arguments to search for, check :class:`pykechain.Client.parts`
                       for additional info
        :type kwargs: dict
        :return: a set of `Parts` as a :class:`PartSet`. Will be empty if no siblings.
        :raises APIError: When an error occurs.
        """
        if self.parent_id:
            return self._client.parts(parent=self.parent_id, category=self.category, **kwargs)
        else:
            from pykechain.models.partset import PartSet
            return PartSet(parts=[])

    def model(self):
        """
        Retrieve the model of this `Part` as `Part`.

        For instance, you can get the part model of a part instance. But trying to get the model of a part that
        has no model, like a part model, will raise a :exc:`NotFoundError`.

        .. versionadded:: 1.8

        :return: the model of this part instance as :class:`Part` with category `MODEL`
        :raises NotFoundError: if no model found

        Example
        -------
        >>> front_fork = project.part('Front Fork')
        >>> front_fork_model = front_fork.model()

        """
        if self.category == Category.INSTANCE:
            model_id = self._json_data['model'].get('id')
            return self._client.model(pk=model_id)
        else:
            raise NotFoundError("Part {} has no model".format(self.name))

    def instances(self, **kwargs):
        """
        Retrieve the instances of this `Part` as a `PartSet`.

        For instance, if you have a model part, you can get the list of instances that are created based on this
        moodel. If there are no instances (only possible if the multiplicity is :attr:`enums.Multiplicity.ZERO_MANY`)
        than a :exc:`NotFoundError` is returned

        .. versionadded:: 1.8

        :return: the instances of this part model :class:`PartSet` with category `INSTANCE`
        :raises NotFoundError: if no instances found

        Example
        -------
        >>> wheel_model = project.model('Wheel')
        >>> wheel_instance_set = wheel_model.instances()

        An example with retrieving the front wheels only using the 'name__contains' search argument.

        >>> wheel_model = project.model('Wheel')
        >>> front_wheel_instances = wheel_model.instances(name__contains='Front')

        """
        if self.category == Category.MODEL:
            return self._client.parts(model=self, category=Category.INSTANCE, **kwargs)
        else:
            raise NotFoundError("Part {} is not a model".format(self.name))

    def instance(self):
        """
        Retrieve the single (expected) instance of this 'Part' (of `Category.MODEL`) as a 'Part'.

        See :func:`Part.instances()` method for documentation.

        :return: :class:`Part` with category `INSTANCE`
        :raises NotFoundError: if the instance does not exist
        :raises MultipleFoundError: if there are more than a single instance returned
        """
        instances_list = list(self.instances())
        if len(instances_list) == 1:
            return instances_list[0]
        elif len(instances_list) > 1:
            raise MultipleFoundError("Part {} has more than a single instance. "
                                     "Use the `Part.instances()` method".format(self.name))
        else:
            raise NotFoundError("Part {} has no instance".format(self.name))

    def proxy_model(self):
        """
        Retrieve the proxy model of this proxied `Part` as a `Part`.

        Allows you to retrieve the model of a proxy. But trying to get the catalog model of a part that
        has no proxy, will raise an :exc:`NotFoundError`. Only models can have a proxy.

        :return: :class:`Part` with category `MODEL` and from which the current part is proxied
        :raises NotFoundError: When no proxy model is found

        Example
        -------
        >>> proxy_part = project.model('Proxy based on catalog model')
        >>> catalog_model_of_proxy_part = proxy_part.proxy_model()

        >>> proxied_material_of_the_bolt_model = project.model('Bolt Material')
        >>> proxy_basis_for_the_material_model = proxied_material_of_the_bolt_model.proxy_model()

        """
        if self.category != Category.MODEL:
            raise IllegalArgumentError("Part {} is not a model, therefore it cannot have a proxy model".format(self))
        if 'proxy' in self._json_data and self._json_data.get('proxy'):
            catalog_model_id = self._json_data['proxy'].get('id')
            return self._client.model(pk=catalog_model_id)
        else:
            raise NotFoundError("Part {} is not a proxy".format(self.name))

    def add(self, model, **kwargs):
        # type: (Part, **Any) -> Part
        """Add a new child instance, based on a model, to this part.

        This can only act on instances. It needs a model from which to create the child instance.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :type kwargs: dict or None
        :type model: :class:`Part`
        :param kwargs: (optional) additional keyword=value arguments
        :type kwargs: dict
        :return: :class:`Part` with category `INSTANCE`.
        :raises APIError: if unable to add the new child instance

        Example
        -------
        >>> bike = project.part('Bike')
        >>> wheel_model = project.model('Wheel')
        >>> bike.add(wheel_model)

        """
        if self.category != Category.INSTANCE:
            raise APIError("Part should be of category INSTANCE")

        return self._client.create_part(self, model, **kwargs)

    def add_to(self, parent, **kwargs):
        # type: (Part, **Any) -> Part
        """Add a new instance of this model to a part.

        This works if the current part is a model and an instance of this model is to be added
        to a part instances in the tree.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param parent: part to add the new instance to
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :type kwargs: dict or None
        :type parent: :class:`Part`
        :param kwargs: (optional) additional keyword=value arguments
        :type kwargs: dict
        :return: :class:`Part` with category `INSTANCE`
        :raises APIError: if unable to add the new child instance

        Example
        -------
        >>> wheel_model = project.model('wheel')
        >>> bike = project.part('Bike')
        >>> wheel_model.add_to(bike)

        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")

        return self._client.create_part(parent, self, **kwargs)

    def add_model(self, *args, **kwargs):
        # type: (*Any, **Any) -> Part
        """Add a new child model to this model.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :return: a :class:`Part` of category `MODEL`
        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")

        return self._client.create_model(self, *args, **kwargs)

    def add_proxy_to(self, parent, name, multiplicity=Multiplicity.ONE_MANY, **kwargs):
        # type: (Any, AnyStr, Any, **Any) -> Part
        """Add this model as a proxy to another parent model.

        This will add the current model as a proxy model to another parent model. It ensure that it will copy the
        whole subassembly to the 'parent' model.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param name: Name of the new proxy model
        :type name: basestring
        :param parent: parent of the to be proxied model
        :type parent: :class:`Part`
        :param multiplicity: the multiplicity of the new proxy model (default ONE_MANY)
        :type multiplicity: basestring or None
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :type kwargs: dict or None
        :return: the new proxied :class:`Part`.
        :raises APIError: in case an Error occurs

        Examples
        --------
        >>> from pykechain.enums import Multiplicity
        >>> bike_model = project.model('Bike')
        # find the catalog model container, the highest parent to create catalog models under
        >>> catalog_model_container = project.model('Catalog container')
        >>> new_wheel_model = project.create_model(catalog_model_container, 'Wheel Catalog',
        ...                                        multiplicity=Multiplicity.ZERO_MANY)
        >>> new_wheel_model.add_proxy_to(bike_model, "Wheel", multiplicity=Multiplicity.ONE_MANY)

        """
        return self._client.create_proxy_model(self, parent, name, multiplicity, **kwargs)

    def add_property(self, *args, **kwargs):
        # type: (*Any, **Any) -> Property
        """Add a new property to this model.

        See :class:`pykechain.Client.create_property` for available parameters.

        :return: :class:`Property`
        :raises APIError: in case an Error occurs
        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")

        return self._client.create_property(self, *args, **kwargs)

    def delete(self):
        # type: () -> None
        """Delete this part.

        :return: None
        :raises APIError: in case an Error occurs
        """
        response = self._client._request('DELETE', self._client._build_url('part', part_id=self.id))

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError("Could not delete part: {} with id {}".format(self.name, self.id))

    def edit(self, name=None, description=None, **kwargs):
        # type: (AnyStr, AnyStr, **Any) -> None
        """
        Edit the details of a part (model or instance).

        For an instance you can edit the Part instance name and the part instance description. To alter the values
        of properties use :func:`Part.update()`.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param name: optional name of the part to edit
        :param description: (optional) description of the part
        :type description: basestring or None
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :type kwargs: dict or None
        :return: the updated object if successful
        :raises IllegalArgumentError: when the type or value of an argument provided is incorrect
        :raises APIError: in case an Error occurs

        Example
        -------
        For changing a part:

        >>> front_fork = project.part('Front Fork')
        >>> front_fork.edit(name='Front Fork - updated')
        >>> front_fork.edit(name='Front Fork cruizer', description='With my ragtop down so my hair can blow' )

        for changing a model:

        >>> front_fork = project.model('Front Fork')
        >>> front_fork.edit(name='Front Fork basemodel', description='Some description here')

        """
        update_dict = {'id': self.id}
        if name is not None:
            if not isinstance(name, (string_types, text_type)):
                raise IllegalArgumentError("name should be provided as a string")
            update_dict.update({'name': name})
        if description is not None:
            if not isinstance(description, (string_types, text_type)):
                raise IllegalArgumentError("description should be provided as a string")
            update_dict.update({'description': description})

        if kwargs is not None:  # pragma: no cover
            update_dict.update(**kwargs)
        response = self._client._request('PUT', self._client._build_url('part', part_id=self.id), json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Part ({})".format(response))

        if name:
            self.name = name

    def _repr_html_(self):
        # type: () -> str

        html = [
            "<table width=100%>",
            "<caption>{}</caption>".format(self.name),
            "<tr>",
            "<th>Property</th>",
            "<th>Value</th>",
            "</tr>"
        ]

        for prop in self.properties:
            style = "color:blue;" if prop._json_data.get('output', False) else ""

            html.append("<tr style=\"{}\">".format(style))
            html.append("<td>{}</td>".format(prop.name))
            html.append("<td>{}</td>".format(prop.value))
            html.append("</tr>")

        html.append("</table>")

        return ''.join(html)

    def update(self, name=None, update_dict=None, bulk=True, **kwargs):
        """
        Edit part name and property values in one go.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param name: new part name (defined as a string)
        :type name: basestring or None
        :param update_dict: dictionary with keys being property names (str) or property ids (uuid)
                            and values being property values
        :type update_dict: dict
        :param bulk: True to use the bulk_update_properties API endpoint for KE-chain versions later then 2.1.0b
        :type bulk: boolean or None
        :param kwargs: (optional) additional keyword arguments that will be passed inside the update request
        :type kwargs: dict or None
        :return: the updated :class:`Part`
        :raises NotFoundError: when the property name is not a valid property of this part
        :raises IllegalArgumentError: when the type or value of an argument provided is incorrect
        :raises APIError: in case an Error occurs

        Example
        -------
        >>> bike = client.scope('Bike Project').part('Bike')
        >>> bike.update(name='Good name', update_dict={'Gears': 11, 'Total Height': 56.3}, bulk=True)

        """
        # dict(name=name, properties=json.dumps(update_dict))) with property ids:value
        action = 'bulk_update_properties'

        request_body = dict()
        for prop_name_or_id, property_value in update_dict.items():
            if isinstance(property_value, Base):
                # is the value is a reference property to another 'Base' Part.
                property_value = property_value.id
            if is_uuid(prop_name_or_id):
                request_body[prop_name_or_id] = property_value
            else:
                request_body[self.property(prop_name_or_id).id] = property_value

        if bulk and len(update_dict.keys()) > 1:
            if name:
                if not isinstance(name, (string_types, text_type)):
                    raise IllegalArgumentError("Name of the part should be provided as a string")
            response = self._client._request('PUT',
                                             self._client._build_url('part', part_id=self.id),
                                             params=dict(select_action=action),
                                             data=dict(name=name, properties=json.dumps(request_body), **kwargs))

            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise APIError("Could not update the part '{}', got: '{}'".format(self, response.content))

            # update local properties (without a call)
            self._json_data = response.json()['results'][0]
            self.properties = [Property.create(p, client=self._client) for p in self._json_data['properties']]
        else:
            for property_name, property_value in update_dict.items():
                self.property(property_name).value = property_value

    def add_with_properties(self, model, name=None, update_dict=None, bulk=True, **kwargs):
        """
        Add a part and update its properties in one go.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param model: model of the part which to add a new instance, should follow the model tree in KE-chain
        :type model: :class:`Part`
        :param name: (optional) name provided for the new instance as string otherwise use the name of the model
        :type name: basestring or None
        :param update_dict: dictionary with keys being property names (str) or property_id (from the property models)
                            and values being property values
        :type update_dict: dict or None
        :param bulk: True to use the bulk_update_properties API endpoint for KE-chain versions later then 2.1.0b
        :type bulk: boolean or None
        :param kwargs: (optional) additional keyword arguments that will be passed inside the update request
        :type kwargs: dict or None
        :return: the newly created :class:`Part`
        :raises NotFoundError: when the property name is not a valid property of this part
        :raises APIError: in case an Error occurs

        Examples
        --------
        >>> bike = client.scope('Bike Project').part('Bike')
        >>> wheel_model = client.scope('Bike Project').model('Wheel')
        >>> bike.add_with_properties(wheel_model, 'Wooden Wheel', {'Spokes': 11, 'Material': 'Wood'})

        """
        if self.category != Category.INSTANCE:
            raise APIError("Part should be of category INSTANCE")
        name = name or model.name
        action = 'new_instance_with_properties'

        properties_update_dict = dict()
        for prop_name_or_id, property_value in update_dict.items():
            if isinstance(property_value, Base):
                # is the value is a reference property to another 'Base' Part.
                property_value = property_value.id
            if is_uuid(prop_name_or_id):
                properties_update_dict[prop_name_or_id] = property_value
            else:
                properties_update_dict[model.property(prop_name_or_id).id] = property_value

        if bulk:
            response = self._client._request('POST', self._client._build_url('parts'),
                                             data=dict(
                                                 name=name,
                                                 model=model.id,
                                                 parent=self.id,
                                                 properties=json.dumps(properties_update_dict),
                                                 **kwargs),
                                             params=dict(select_action=action))

            if response.status_code != requests.codes.created:  # pragma: no cover
                raise APIError('{}: {}'.format(str(response), response.content))
            return Part(response.json()['results'][0], client=self._client)
        else:  # do the old way
            new_part = self.add(model, name=name)  # type: Part
            new_part.update(update_dict=update_dict, bulk=bulk)
            return new_part

    def as_dict(self):
        """
        Retrieve the properties of a part inside a dict in this structure: {property_name: property_value}.

        .. versionadded:: 1.9

        :returns: the values of the properties as a `dict`
        :rtype: dict

        Example
        -------
        >>> front_wheel = client.scope('Bike Project').part('Front Wheel')
        >>> front_wheel_properties = front_wheel.as_dict()
        {'Diameter': 60.8,
         'Spokes': 24,
         'Rim Material': 'Aluminium',
         'Tire Thickness': 4.2}

        """
        properties_dict = dict()
        for prop in self.properties:
            properties_dict[prop.name] = prop.value
        return properties_dict

    def order_properties(self, property_list=None):
        """
        Order the properties of a part model using a list of property objects or property names or property id's.

        :param property_list: ordered list of property names (basestring) or property id's (uuid)
        :type property_list: list(basestring)
        :returns: the :class:`Part` with the reordered list of properties
        :raises APIError: when an Error occurs
        :raises IllegalArgumentError: When provided a wrong argument

        Examples
        --------
        >>> front_fork = client.scope('Bike Project').model('Front Fork')
        >>> front_fork.order_properties(['Material', 'Height (mm)', 'Color'])

        >>> front_fork = client.scope('Bike Project').model('Front Fork')
        >>> material = front_fork.property('Material')
        >>> height = front_fork.property('Height (mm)')
        >>> color = front_fork.property('Color')
        >>> front_fork.order_properties([material, height, color])

        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")
        if not isinstance(property_list, list):
            raise IllegalArgumentError('Expected a list of strings or Property() objects, got a {} object'.
                                       format(type(property_list)))

        order_dict = dict()

        for prop in property_list:
            if isinstance(prop, (str, text_type)):
                order_dict[self.property(name=prop).id] = property_list.index(prop)
            else:
                order_dict[prop.id] = property_list.index(prop)

        response = self._client._request('PUT', self._client._build_url('part', part_id=self.id),
                                         data=dict(property_order=json.dumps(order_dict)))
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not reorder properties")

    def populate_descendants(self, batch=200):
        """
        Retrieve the descendants of a specific part in a list of dicts and populate the :func:`Part.children()` method.

        Each `Part` has a :func:`Part.children()` method to retrieve the children on the go. This function
        prepopulates the children and the children's children with its children in one call, making the traversal
        through the parttree blazingly fast.

        .. versionadded:: 2.1

        :param batch: Number of Parts to be retrieved in a batch
        :type batch: int (defaults to 200)
        :returns: list of `Parts` with `children`
        :raises APIError: if you cannot create the children tree.

        Example
        -------
        >>> bike = client.part('Bike')
        >>> bike.populate_descendants(batch=150)

        """
        descendants_flat_list = list(self._client.parts(
            parent_id=self.id,
            category=self.category,
            batch=batch)
        )

        self._cached_children = descendants_flat_list

    def clone(self, **kwargs):
        """
        Clone a part.

        An optional name of the cloned part may be provided. If not provided the name will be set
        to "CLONE - <part name>". (available for KE-chain 3 backends only)
        An optional multiplicity, may be added as paremeter for the cloning of models. If not
        provided the multiplicity of the part will be used. (available for KE-chain 3 backends only)

        .. versionadded:: 2.3
        .. versionchanged:: 3.0
           Added optional paramenters the name and multiplicity for KE-chain 3 backends.

        :param kwargs: (optional) additional keyword=value arguments
        :type kwargs: dict
        :return: cloned :class:`models.Part`
        :raises APIError: if the `Part` could not be cloned

        Example
        -------
        >>> bike = project.model('Bike')
        >>> bike2 = bike.clone()

        For KE-chain 3 backends

        >>> bike = project.model('Bike')
        >>> bike2 = bike.clone(name='Trike', multiplicity=Multiplicity.ZERO_MANY)

        """
        parent = self.parent()
        return self._client._create_clone(parent, self, **kwargs)

    def copy(self, target_parent, name=None, include_children=True, include_instances=True):
        """
        Copy the `Part` to target parent, both of them having the same category.

        .. versionadded:: 2.3

        :param target_parent: `Part` object under which the desired `Part` is copied
        :type target_parent: :class:`Part`
        :param name: how the copied top-level `Part` should be called
        :type name: basestring
        :param include_children: True to copy also the descendants of `Part`.
        :type include_children: bool
        :param include_instances: True to copy also the instances of `Part` to ALL the instances of target_parent.
        :type include_instances: bool
        :returns: copied :class:`Part` model.
        :raises IllegalArgumentError: if part and target_parent have different `Category`
        :raises IllegalArgumentError: if part and target_parent are identical

        Example
        -------
        >>> model_to_copy = client.model(name='Model to be copied')
        >>> bike = client.model('Bike')
        >>> model_to_copy.copy(target_parent=bike, name='Copied model',
        >>>                    include_children=True,
        >>>                    include_instances=True)

        """
        if not isinstance(target_parent, Part):
            raise IllegalArgumentError("`target_parent` needs to be a part, got '{}'".format(type(target_parent)))

        if self.category == Category.MODEL and target_parent.category == Category.MODEL:
            # Cannot add a model under an instance or vice versa
            copied_model = relocate_model(part=self, target_parent=target_parent, name=name,
                                          include_children=include_children)
            if include_instances:
                instances_to_be_copied = list(self.instances())
                parent_instances = list(target_parent.instances())
                for parent_instance in parent_instances:
                    for instance in instances_to_be_copied:
                        instance.populate_descendants()
                        move_part_instance(part_instance=instance, target_parent=parent_instance,
                                           part_model=self, name=instance.name,
                                           include_children=include_children)
            return copied_model

        elif self.category == Category.INSTANCE and target_parent.category == Category.INSTANCE:
            copied_instance = relocate_instance(part=self, target_parent=target_parent, name=name,
                                                include_children=include_children)
            return copied_instance
        else:
            raise IllegalArgumentError('part "{}" and target parent "{}" must have the same category')

    def move(self, target_parent, name=None, include_children=True, include_instances=True):
        """
        Move the `Part` to target parent, both of them the same category.

        .. versionadded:: 2.3

        :param target_parent: `Part` object under which the desired `Part` is moved
        :type target_parent: :class:`Part`
        :param name: how the moved top-level `Part` should be called
        :type name: basestring
        :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
        :type include_children: bool
        :param include_instances: True to move also the instances of `Part` to ALL the instances of target_parent.
        :type include_instances: bool
        :returns: moved :class:`Part` model.
        :raises IllegalArgumentError: if part and target_parent have different `Category`
        :raises IllegalArgumentError: if target_parent is descendant of part

        Example
        -------
        >>> model_to_move = client.model(name='Model to be moved')
        >>> bike = client.model('Bike')
        >>> model_to_move.move(target_parent=bike, name='Moved model',
        >>>                    include_children=True,
        >>>                    include_instances=True)

        """
        if not name:
            name = self.name
        if self.category == Category.MODEL and target_parent.category == Category.MODEL:
            moved_model = relocate_model(part=self, target_parent=target_parent, name=name,
                                         include_children=include_children)
            if include_instances:
                retrieve_instances_to_copied = list(self.instances())
                retrieve_parent_instances = list(target_parent.instances())
                for parent_instance in retrieve_parent_instances:
                    for instance in retrieve_instances_to_copied:
                        instance.populate_descendants()
                        move_part_instance(part_instance=instance, target_parent=parent_instance,
                                           part_model=self, name=instance.name, include_children=include_children)
            self.delete()
            return moved_model
        elif self.category == Category.INSTANCE and target_parent.category == Category.INSTANCE:
            moved_instance = relocate_instance(part=self, target_parent=target_parent, name=name,
                                               include_children=include_children)
            try:
                self.delete()
            except APIError:
                model_of_instance = self.model()
                model_of_instance.delete()
            return moved_instance
        else:
            raise IllegalArgumentError('part "{}" and target parent "{}" must have the same category')



import warnings
from typing import Dict, Optional, Union, Text, Tuple, List

from six import text_type

from pykechain.enums import Category, PropertyType, WidgetTitleValue
from pykechain.exceptions import IllegalArgumentError
from pykechain.utils import is_uuid, snakecase, camelcase

# these are the common keys to all kecards.
KECARD_COMMON_KEYS = ["collapsed", "collapsible", "noBackground", "noPadding", "isDisabled", "isMerged"]


def _retrieve_object(obj: Union['Part2', 'AnyProperty', 'Team', 'Service', Text],
                     method: 'Client',
                     ) -> Union['Part2', 'AnyProperty', 'Team', 'Service']:
    """
    Object if object or uuid of object is provided as argument.

    :param obj: object or uuid to retrieve the object for
    :type obj: :class:`Base` or basestring
    :param method: client object to retrieve the object if only uuid is provided.
    :type method: `Client`
    :return: object based on the object or uuid of the objet
    :rtype: `Part2` or `Team` or `Property2`
    :raises APIError: If the object could not be retrieved based on the UUID
    :raises IllegalArgumentError: if the object provided is not a Part, Property2 or UUID.
    """
    # Check whether the part_model is uuid type or class `Part`
    from pykechain.models import Part, Part2, Property, Property2, Service, Team
    if isinstance(obj, (Part, Part2, Property, Property2, Service, Team)):
        return obj
    elif isinstance(obj, text_type) and is_uuid(obj):
        obj_id = obj
        obj = method(id=obj_id)
        return obj
    else:
        raise IllegalArgumentError("When adding the widget, obj must be a Part, Property, Service, Team, "
                                   " Part id, Property id, Service id or Team id. Type is: {}".format(type(obj)))


def _retrieve_object_id(obj: Optional[Union['Base', Text]]) -> Optional[Text]:
    """
    Object id if object or uuid of object is provided as argument.

    :param obj: object or uuid to retrieve the object id for
    :type obj: :class:`Base` or basestring
    :return: object based on the object or uuid of the object
    :rtype: basestring or None
    :raises APIError: If the object could not be retrieved based on the UUID
    :raises IllegalArgumentError: if the object provided is not a Base instance or UUID.
    """
    # Check whether the obj is an object of any subclass of Base, or uuid type
    from pykechain.models import Base
    if issubclass(type(obj), Base):
        return obj.id
    elif isinstance(obj, text_type) and is_uuid(obj):
        return obj
    elif isinstance(obj, type(None)):
        return None
    else:
        raise IllegalArgumentError("When adding the widget, obj must be an instance of `Base` or an object id. "
                                   "Type is: {}".format(type(obj)))


def _set_title(meta: Dict,
               title: Optional[Union[Text, bool]] = None,
               custom_title: Optional[Union[Text, bool]] = None,
               default_title: Optional[Text] = None,
               show_title_value: Optional[WidgetTitleValue] = None,
               **kwargs) -> Tuple[Dict, Text]:
    """
    Set the customTitle in the meta based on provided optional custom title or default.

    This will inject into the meta the `customTitle` and `showTitleValue` if the custom_title is provided as
    argument, otherwise it will inject the `defaultTitle`. It returns the meta definition of the widget and the
    title of the widget (to be used to set `widget.title`).

    :param meta: meta dictionary to augment
    :type meta: dict
    :param custom_title: A custom title for the multi column widget
            * False: use the default title
            * String value: use the Custom title
            * None: No title at all.
    :type custom_title: basestring or bool or None
    :param default_title: (optional) If custom_title is False, the default title is injected as title
    :type default_title: basestring or None
    :param show_title_value: (optional) Specify how the title is displayed, regardless of other inputs.
    :type show_title_value: WidgetTitleValue
    :return: tuple of meta and the title
    :rtype: Tuple[Dict,Text]
    :raises IllegalArgumentError: When illegal (combination) of arguments are set.
    """
    if custom_title and title is None:
        warnings.warn('`custom_title` attribute will be deprecated in version 3.4.0, please adapt your code '
                      'accordingly to use `title`.', PendingDeprecationWarning)
        title = custom_title

    if show_title_value is None:
        if title is False:
            show_title_value = WidgetTitleValue.DEFAULT
        elif title is None or title == '':
            show_title_value = WidgetTitleValue.NO_TITLE
        else:
            show_title_value = WidgetTitleValue.CUSTOM_TITLE

    elif show_title_value not in WidgetTitleValue.values():
        raise IllegalArgumentError('`show_title_value` must be a WidgetTitleValue enum option, "{}" is not. '
                                   'Choose from: {}'.format(show_title_value, WidgetTitleValue.values()))

    if show_title_value == WidgetTitleValue.DEFAULT:
        title_meta = None
        if default_title is None:
            raise IllegalArgumentError("If `title` is set to False the `default_title` is used and cannot be None. "
                                       "Provide a `default_title` argument and ensure it is not None.")
        title_ref = default_title
    else:
        title_meta = title
        title_ref = title

    meta.update({
        "showTitleValue": show_title_value,
        "customTitle": title_meta,
    })
    return meta, title_ref


def _initiate_meta(kwargs, activity, ignores=()):
    """
    Init Widget meta definition.

    Mainly ensure to initiate the correct keys that are common to (most) widgets like the keys related
    to the so called kecard (the card object around the widget in the frontend. These keys are: `collapsed`,
    `collapsible`, `noBackground`, `noPadding`, `isDisabled`, `customHeight`. Also the
    the `custom_height` is initiated.

    If you want to be sure that some keys are never present in the returned meta (eg 'customHeight')
    than you may want to add them to ignores.

    :param kwargs: the keyword arguments provided to the widget function which are checked for the kecard definition
    :type kwargs: dict
    :param activity: uuid or `activity2` object
    :type activity: Activity2 or basestring
    :param ignores: list or tuple of keys to ensure they are not present in the initated meta on return
    :type ignores: list or tuple
    :return: kecard dictionary
    :rtype: dict
    """
    meta = dict(activityId=str(_retrieve_object_id(activity)))
    # also add the keys' in their snake case appearance so noPadding and no_padding, customHeight and custom_height
    keys_in_kwargs = KECARD_COMMON_KEYS + [snakecase(k) for k in KECARD_COMMON_KEYS]

    # initiate the meta based on known kwarg arguments
    for key in list(set(keys_in_kwargs)):
        if key in kwargs:
            meta[camelcase(key)] = kwargs.pop(key)

    # we check for custom_height specifically and deal with it.
    if snakecase('customHeight') in kwargs:
        meta['customHeight'] = kwargs.pop(snakecase("customHeight"))

    # remove the 'ignores' from the meta
    for key in ignores:
        if key in meta:
            del meta[key]

    return meta


def _check_prefilters(part_model: 'Part2', prefilters: Dict) -> List[Text]:  # noqa: F821
    """
    Check the pre-filters on a `FilteredGridWidget` or `MultiReferenceProperty2.

    The prefilters should comply to the prefilters schema as present in the backend.

    ```
    "prefilters": {
        "type": ["object", "null"],
        "additionalProperties": True,
        "properties": {
            "property_value": def_nullstring,
            "name__icontains": def_nullstring
        }
    }
    ```

    :param part_model: The part model of which the properties are filtered.
    :param prefilters: Dictionary with prefilters.
    :raises IllegalArgumentError: when the type of the input is provided incorrect.
    """
    from pykechain.models import Property2

    property_models = prefilters.get('property_models', [])  # type: List[Property2, Text]  # noqa
    values = prefilters.get('values', [])
    filters_type = prefilters.get('filters_type', [])

    if any(len(lst) != len(property_models) for lst in [values, filters_type]):
        raise IllegalArgumentError('The lists of "property_models", "values" and "filters_type" should be the '
                                   'same length.')

    list_of_prefilters = []
    for prop, value, filter_type in zip(property_models, values, filters_type):
        if is_uuid(prop):
            prop = part_model.property(prop)  # the finder can handle uuid, name or ref

        if not isinstance(prop, Property2) or prop.category != Category.MODEL:
            raise IllegalArgumentError(
                'Pre-filters can only be set on Property models, received "{}".'.format(prop))

        # TODO when a property is freshly created, the property has no "part_id" key in json_data.
        elif part_model.id != prop.part_id:
            raise IllegalArgumentError(
                'Pre-filters can only be set on properties belonging to the selected Part model, found '
                'selected Part model "{}" and Properties belonging to "{}"'.format(part_model.name,
                                                                                   prop.part.name))
        else:
            if prop.type == PropertyType.BOOLEAN_VALUE:
                value = str(value).lower()
            new_pre_filter = ':'.join([prop.id, str(value), filter_type])
            list_of_prefilters.append(new_pre_filter)
    return list_of_prefilters


def _check_excluded_propmodels(part_model: 'Part2', property_models: List['AnyProperty']) -> List['AnyProperty']:
    """
    Validate the excluded property models of the referenced part.

    :param part_model: Part model that is referenced
    :param property_models: Properties of the part model to be excluded.
    :return: list of property IDs
    :rtype list
    :raises IllegalArgumentError: whenever the properties are not of category MODEL or do not belong to the part model.
    """
    from pykechain.models import Property2
    from pykechain.models import Part2

    if not isinstance(part_model, Part2):
        raise IllegalArgumentError('`part_model` must be a Part2 object, "{}" is not.'.format(part_model))

    list_of_propmodels_excl = list()  # type: List['AnyProperty']
    for property_model in property_models:
        if is_uuid(property_model):
            property_model = part_model.property(property_model)
        elif not isinstance(property_model, Property2):
            raise IllegalArgumentError('A part reference property can only exclude `Property` models or their UUIDs, '
                                       'found type "{}"'.format(type(property_model)))

        if property_model.category != Category.MODEL:
            raise IllegalArgumentError('A part reference property can only exclude `Property` models, found '
                                       'category "{}" on property "{}"'.format(property_model.category,
                                                                               property_model.name))
        # TODO when a property is freshly created, the property has no "part_id" key in json_data.
        elif part_model.id != property_model.part_id:
            raise IllegalArgumentError(
                'A part reference property can only exclude properties belonging to the referenced Part model, '
                'found referenced Part model "{}" and Properties belonging to "{}"'.format(
                    part_model.name, property_model.part.name))
        else:
            list_of_propmodels_excl.append(property_model.id)
    return list_of_propmodels_excl



from unittest import TestCase

from pykechain.enums import Enum


class FirstEnum(Enum):
    SLUG_1 = 'first slug'


class SecondEnum(FirstEnum):
    SLUG_2 = 'second slug'


class TestEnums(TestCase):

    def test_inheritance(self):
        values = SecondEnum.values()

        self.assertIn('first slug', values)



"""All validators and effect objects.

.. versionadded:: 2.2
"""

from .effects import (
    TextEffect,
    ErrorTextEffect,
    HelpTextEffect,
    VisualEffect,
    ValidVisualEffect,
    InvalidVisualEffect
)
from .validators import (
    NumericRangeValidator,
    BooleanFieldValidator,
    RequiredFieldValidator,
    OddNumberValidator,
    EvenNumberValidator,
    RegexStringValidator,
    SingleReferenceValidator
)
from .validators_base import PropertyValidator, ValidatorEffect  # noqa

__all__ = (
    'TextEffect', 'ErrorTextEffect', 'HelpTextEffect', 'VisualEffect', 'ValidVisualEffect', 'InvalidVisualEffect',
    'NumericRangeValidator', 'BooleanFieldValidator', 'RequiredFieldValidator', 'OddNumberValidator',
    'EvenNumberValidator', 'RegexStringValidator', 'SingleReferenceValidator'
)



import datetime
from time import sleep
from unittest import skipIf

from pykechain.enums import ScopeStatus, KEChainPages
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models import Team, Scope2
from pykechain.models.sidebar.sidebar_manager import SideBarManager
from tests.classes import TestBetamax
from tests.utils import TEST_FLAG_IS_PIM3


class TestScopes(TestBetamax):

    def test_retrieve_scopes(self):
        self.assertTrue(self.client.scopes())

    def test_scope_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'name', 'created_at', 'updated_at', 'ref',
                      'process', 'workflow_root', 'description', 'status', 'category', 'tags',
                      'start_date', 'due_date']

        obj = self.project
        for attribute in attributes:
            self.assertTrue(hasattr(obj, attribute),
                            "Could not find '{}' in the object: '{}'".format(attribute, obj.__dict__.keys()))

    def test_retrieve_scope_with_kwargs(self):
        retrieve_scopes_with_kwargs = self.client.scopes(name__icontains='Bike')
        self.assertTrue(retrieve_scopes_with_kwargs)
        self.assertTrue(len(retrieve_scopes_with_kwargs) >= 1)

    def test_retrieve_single_unknown(self):
        with self.assertRaises(NotFoundError):
            self.client.scope('123lladadwd')

    def test_retrieve_single_multiple(self):
        with self.assertRaises(MultipleFoundError):
            self.client.scope()

    def test_add_member(self):
        member_to_be_added = 'anotheruser'
        # testing
        self.project.add_member(member_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_members = self.project.members()
        self.assertTrue(member_to_be_added in [member['username'] for member in project_members])
        # teardown
        self.project.remove_member(member_to_be_added)

    def test_add_member_by_id(self):
        member_id = 4
        with self.assertRaises(TypeError):
            self.project.add_member(member_id)

    def test_add_non_existing_member(self):
        member_to_be_added = 'Nobody'
        with self.assertRaises(NotFoundError):
            self.project.add_member(member_to_be_added)

    def test_remove_member(self):
        member_to_be_removed = 'anotheruser'
        # setUp
        self.project.add_member(member_to_be_removed)
        # testing
        self.project.remove_member(member_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_members = self.project.members()
        self.assertTrue(member_to_be_removed not in [member['username'] for member in project_members])

    def test_add_manager(self):
        manager_to_be_added = 'anotheruser'
        # testing
        self.project.add_manager(manager_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_managers = self.project.members(is_manager=True)
        self.assertTrue(manager_to_be_added in [manager['username'] for manager in project_managers])
        # teardown
        self.project.remove_member(manager_to_be_added)

    def test_remove_manager(self):
        manager_to_be_removed = 'anotheruser'
        # setUp
        self.project.add_manager(manager_to_be_removed)
        # testing
        self.project.remove_manager(manager_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_managers = self.project.members(is_manager=True)
        self.assertTrue(manager_to_be_removed not in [manager['username'] for manager in project_managers])
        project_managers = self.project.members(is_manager=False)
        self.assertTrue(manager_to_be_removed in [manager['username'] for manager in project_managers])
        # teardown
        self.project.remove_member(manager_to_be_removed)

    def test_add_leadmember(self):
        leadmember_to_be_added = 'anotheruser'
        # testing
        self.project.add_leadmember(leadmember_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_leadmembers = self.project.members(is_leadmember=True)
        self.assertTrue(leadmember_to_be_added in [leadmember['username'] for leadmember in project_leadmembers])
        # teardown
        self.project.remove_member(leadmember_to_be_added)

    def test_remove_leadmember(self):
        leadmember_to_be_removed = 'anotheruser'
        # setUp
        self.project.add_leadmember(leadmember_to_be_removed)
        # testing
        self.project.remove_leadmember(leadmember_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_leadmembers = self.project.members(is_leadmember=True)
        self.assertTrue(leadmember_to_be_removed not in [leadmember['username'] for leadmember in project_leadmembers])
        project_leadmembers = self.project.members(is_leadmember=False)
        self.assertTrue(leadmember_to_be_removed in [leadmember['username'] for leadmember in project_leadmembers])
        # teardown
        self.project.remove_member(leadmember_to_be_removed)

    # 2.2.0+
    def test_edit_scope(self):
        # setUp
        new_scope_name = 'Pykechain testing (bike project)'
        old_scope_name = self.project.name
        new_scope_description = 'Project used to build a Bike. Part-time job is to also test Pykechain.'
        old_scope_description = 'Project used to test Pykechain and push it to the limits'
        new_start_date = datetime.datetime(2018, 12, 5, tzinfo=None)
        old_start_date = datetime.datetime(2017, 4, 12)
        new_due_date = datetime.datetime(2018, 12, 8, tzinfo=None)
        old_due_date = datetime.datetime(2018, 4, 1)
        old_status = ScopeStatus.ACTIVE

        self.project.edit(name=new_scope_name, description=new_scope_description, start_date=new_start_date,
                          due_date=new_due_date, status=ScopeStatus.CLOSED)

        # testing
        retrieved_project = self.client.scope(id=self.project.id, status=ScopeStatus.CLOSED)

        self.assertTrue(retrieved_project.name == new_scope_name)
        self.assertTrue(retrieved_project._json_data['text'] == new_scope_description)
        self.assertTrue(retrieved_project._json_data['start_date'] in ('2018-12-05T00:00:00Z', '2018-12-05T00:00:00+00:00'))
        self.assertTrue(retrieved_project._json_data['due_date'] in ('2018-12-08T00:00:00Z', '2018-12-08T00:00:00+00:00'))
        self.assertTrue(retrieved_project._json_data['status'] == ScopeStatus.CLOSED)

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(name=True)

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(description=new_start_date)

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(start_date='Yes, man!')

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(due_date='Wrong')

        with self.assertRaises(IllegalArgumentError):
            self.project.edit(status='ILLEGAL_STATUS_WILL_CAUSE_ERR')

        # tearDown
        self.project.edit(name=old_scope_name, description=old_scope_description, start_date=old_start_date,
                          due_date=old_due_date, status=old_status)

    # v2.3.3
    def test_edit_scope_team(self):
        """Test capabilities of changing team scope."""

        # setup
        team_url = self.project._client._build_url('teams')

        r = self.project._client._request('get', team_url, params=dict(name='Team No.1'))
        team_kew_id = r.json().get('results')[0].get('id')  # no iffer as it should fail hard when not a respcode =ok

        r = self.project._client._request('get', team_url, params=dict(name='Second Team'))
        team_b_id = r.json().get('results')[0].get('id')

        # save current team
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            team_dict = self.project._json_data.get('team_id_name')
            saved_team_id = team_dict and team_dict.get('id')
        else:
            team_dict = self.project._json_data.get('team')
            saved_team_id = team_dict and team_dict.get('id')

        self.project.edit(team=team_b_id)
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            self.assertEqual(self.project._json_data.get('team_id_name').get('id'), team_b_id)
        else:
            self.assertEqual(self.project._json_data.get('team').get('id'), team_b_id)

        self.project.edit(team=team_kew_id)
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            self.assertEqual(self.project._json_data.get('team_id_name').get('id'), team_kew_id)
        else:
            self.assertEqual(self.project._json_data.get('team').get('id'), team_kew_id)

        self.project.edit(team=saved_team_id)
        if self.client.match_app_version(label="pim", version=">=3.0.0"):
            self.assertEqual(self.project._json_data.get('team_id_name').get('id'), saved_team_id)
        else:
            self.assertEqual(self.project._json_data.get('team').get('id'), saved_team_id)

    def test_team_property_of_scope(self):
        """Test for the property 'team' of a scope."""
        team = self.project.team
        self.assertIsInstance(team, Team)


@skipIf(not TEST_FLAG_IS_PIM3, reason="This tests is designed for PIM version 2, expected to fail on old PIM")
class TestScopes2SpecificTests(TestBetamax):

    def setUp(self):
        """
        SetUp the test.

        When you create a scope, assign it to self.scope such that it is deleted when the test is complete

        :ivar scope: scope that you create during a test, which will be autoremoved in the cleanup.
        """
        super(TestScopes2SpecificTests, self).setUp()
        self.scope = None

    def tearDown(self):
        """Teardown the test, remove the scope.

        When you create a scope, assign it to self.scope such that it is deleted when the test is complete
        """
        if self.scope:
            self.scope.delete()
        super(TestScopes2SpecificTests, self).tearDown()

    def test_side_bar(self):
        side_bar_manager = self.project.side_bar()

        self.assertIsInstance(side_bar_manager, SideBarManager)

    def test_set_landing_page_activity(self):
        self.scope = self.project.clone(asynchronous=False)

        first_activity = self.scope.activities()[0]

        self.scope.set_landing_page(activity=first_activity)

    def test_set_landing_page_data_model(self):
        self.scope = self.project.clone(asynchronous=False)

        data_model_page = KEChainPages.DATA_MODEL

        self.scope.set_landing_page(activity=data_model_page)

    def test_retrieve_scope2_members(self):
        original_scope_members = [u.get('username') for u in self.project.members()]
        original_scope_managers = [u.get('username') for u in self.project.members(is_manager=True)]
        scope_managers = original_scope_managers + ['testmanager']
        scope_members = original_scope_members + ['anotheruser', 'testuser']

        self.project.add_manager('testmanager')
        self.project.add_member('anotheruser')
        self.project.add_member('testuser')

        self.project.refresh()

        members_usernames = [member['username'] for member in self.project.members()]
        managers_usernames = [manager['username'] for manager in self.project.members(is_manager=True)]

        self.assertSetEqual(set(scope_members), set(members_usernames))
        self.assertSetEqual(set(scope_managers), set(managers_usernames))

        # teardown
        self.project.remove_member('anotheruser')

    def test_members(self):

        # Helper function:
        def get_member_names(**kwargs):
            return [member['username'] for member in self.project.members(**kwargs)]

        all_members = set(get_member_names())
        managers = set(get_member_names(is_manager=True))
        non_managers = set(get_member_names(is_manager=False))
        leads = set(get_member_names(is_leadmember=True))
        non_leads = set(get_member_names(is_leadmember=False))
        basic_members = set(get_member_names(is_manager=False, is_leadmember=False))
        lead_and_manager = set(get_member_names(is_manager=True, is_leadmember=True))

        self.assertTrue(len(all_members) == 4, msg='Number of default scope members has changed!')
        self.assertTrue(basic_members and leads and managers and lead_and_manager,
                        msg='Test scope must have a member of every type!')

        all_roles = basic_members | leads | managers | lead_and_manager
        remaining_non_leads = all_roles - leads
        remaining_non_managers = all_roles - managers

        self.assertSetEqual(all_roles, all_members, msg='The sum of all roles should equal all members.')
        self.assertSetEqual(remaining_non_leads, non_leads)
        self.assertSetEqual(remaining_non_managers, non_managers)

        self.assertTrue(lead_and_manager <= managers)
        self.assertTrue(lead_and_manager <= leads)

    def test_clone_scope_vanilla(self):
        self.scope = self.project.clone(asynchronous=False)

        self.assertNotEqual(self.project.id, self.scope.id)
        self.assertEqual(self.project._json_data.get('tasks_count'), self.scope._json_data.get('tasks_count'))
        self.assertEqual(self.project._json_data.get('processes_count'), self.scope._json_data.get('processes_count'))
        self.assertEqual(self.project._json_data.get('activities_count'), self.scope._json_data.get('activities_count'))
        self.assertEqual(self.project._json_data.get('parts_count'), self.scope._json_data.get('parts_count'))
        self.assertEqual(self.project._json_data.get('model_parts_count'), self.scope._json_data.get('model_parts_count'))
        self.assertEqual(self.project._json_data.get('properties_count'), self.scope._json_data.get('properties_count'))

    def test_scope_delete(self):
        new_scope = self.project.clone(asynchronous=False)
        self.assertNotEqual(self.project.id, new_scope.id)
        new_scope.delete(asynchronous=False)

        with self.assertRaisesRegex(NotFoundError, 'No scope fits criteria'):
            # throw in arbitrary sleep to give backend time to actually delete the scope.
            sleep(1)
            self.client.scope(pk=new_scope.id)

    def test_clone_scope_updated_name_description_tags_etc(self):
        """
        self, source_scope, name=None, status=None, tags=None, start_date=None, due_date=None,
                    description=None, team=None, scope_options=None, async=False):
        """
        new_scope = self.project.clone(
            name="new bike scope",
            description="new description",
            status=ScopeStatus.CLOSED,
            # scope_options={"hello": "world"},
            tags=('a', 'b', 'b', 'c'),
            asynchronous=False)
        self.scope = new_scope

        self.assertNotEqual(self.project.id, new_scope.id)
        self.assertEqual(new_scope._json_data.get('name'), "new bike scope")
        self.assertEqual(new_scope._json_data.get('text'), "new description")
        self.assertEqual(new_scope._json_data.get('status'), ScopeStatus.CLOSED)
        # self.assertListEqual(new_scope._json_data.get('tags'), ["a", "b", "c"])

    def test_retrieve_scope_with_refs(self):
        # setup
        scope_ref = 'bike-project'
        scope_name = 'Bike Project'
        scope = self.client.scope(ref=scope_ref)

        # testing
        self.assertIsInstance(scope, Scope2)
        self.assertTrue(scope.name, scope_name)



from abc import ABC
from typing import Optional, Text, List

from pykechain.models import Base


class TreeObject(Base, ABC):
    """Object class to include methods used to traverse a tree-structure."""

    def __init__(self, json, **kwargs):
        """
        Initialize the object with attributes related to a tree-structure.

        :param json: data dict from KE-chain.
        """
        super().__init__(json=json, **kwargs)

        self.parent_id = json.get('parent_id', None)  # type: Optional[Text]

        self._cached_children = None  # type: Optional[List['TreeObject']]

    def __call__(self, *args, **kwargs) -> 'TreeObject':
        """Short-hand version of the `child` method."""
        return self.child(*args, **kwargs)

    def child(self,
              name: Optional[Text] = None,
              pk: Optional[Text] = None,
              **kwargs) -> 'TreeObject':
        """
        Retrieve a child object.

        :param name: optional, name of the child
        :type name: str
        :param pk: optional, UUID of the child
        :type: pk: str
        :return: Child object
        :raises MultipleFoundError: whenever multiple children fit match inputs.
        :raises NotFoundError: whenever no child matching the inputs could be found.
        """
        raise NotImplementedError

    def parent(self) -> 'TreeObject':
        """
        Retrieve the parent object.

        :return: Parent object
        :raises APIError: whenever the parent could not be retrieved
        """
        raise NotImplementedError

    def siblings(self, **kwargs) -> List['TreeObject']:
        """
        Retrieve all objects that have the same parent as the current object, thereby including itself.

        :return: iterable of objects
        :raises NotFoundError: whenever the current object is a root object, thereby having no siblings.
        """
        raise NotImplementedError

    def children(self, **kwargs) -> List['TreeObject']:
        """
        Retrieve all children objects.

        :return: iterable of child objects.
        :rtype Iterable
        :raises APIError:
        """
        raise NotImplementedError

    def all_children(self) -> List['TreeObject']:
        """
        Retrieve a flat list of all descendants, sorted depth-first.

        :returns list of child objects
        :rtype List
        """
        all_children = list(self.children())

        for child in self.children():
            all_children.extend(child.all_children())

        return all_children



from abc import abstractmethod
from typing import Dict, Any
from jsonschema import validate

from pykechain.enums import PropertyRepresentation
from pykechain.models.validators.validator_schemas import representation_jsonschema_stub


class BaseRepresentation(object):
    """
    Base class for all Representations.

    :cvar jsonschema: jsonschema to validate the json representation of the Validator
    :type jsonschema: Dict
    :cvar rtype: type of representation
    :type rtype: Text
    """

    jsonschema = representation_jsonschema_stub
    rtype = None
    _config_value_key = None

    def __init__(self, prop, json=None, value=None):
        """Construct a base validator."""
        self._property = prop

        self._json = json or dict(rtype=self.rtype, config=dict())  # type: dict

        self._config = self._json.get('config', dict())  # type: dict

        if value is None and self._config_value_key in self._config:
            self._value = self._config[self._config_value_key]
        else:
            self.validate_representation(value)
            self._config[self._config_value_key] = value
            self._value = value

    def as_json(self):
        # type: () -> dict
        """Parse the validator to a proper validator json."""
        return self._json

    def validate_json(self):
        # type: () -> Any
        """Validate the json representation of the validator against the validator jsonschema."""
        return validate(self._json, self.jsonschema)

    @classmethod
    def parse(cls, prop, json):
        # type: (AnyProperty, Dict) -> BaseRepresentation
        """Parse a json dict and return the correct subclass of :class:`BaseRepresentation`.

        It uses the 'effect' key to determine which :class:`BaseRepresentation` to instantiate.
        Please refer to :class:`pykechain.enums.PropertyRepresentation` for the supported representations.

        :param prop: Property object to which the `BaseRepresentation` belongs.
        :type: prop: AnyProperty
        :param json: dictionary containing the specific keys to parse into a :class:`BaseRepresentation`
        :type json: dict
        :returns: the instantiated subclass of :class:`BaseRepresentation`
        :rtype: :class:`BaseRepresentation` or subclass thereof
        """
        if 'rtype' in json:
            rtype = json.get('rtype')
            if rtype not in PropertyRepresentation.values():
                raise Exception("Representation unknown, incorrect json: '{}'".format(json))

            from pykechain.models.representations import representations
            rtype_implementation_classname = "{}{}".format(rtype[0].upper(), rtype[1:])  # type: ignore
            if hasattr(representations, rtype_implementation_classname):
                return getattr(representations, rtype_implementation_classname)(prop=prop, json=json)
            else:
                raise Exception('unknown rtype in json')
        raise Exception("Representation unknown, incorrect json: '{}'".format(json))

    @property
    def value(self):
        """
        Retrieve current representation value.

        :return: value
        :rtype Any
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Set a new representation value.

        :param value: the new value to be set
        :type value: Any
        :return: the value
        :rtype Any
        """
        self.validate_representation(value)

        self._value = value
        self._config[self._config_value_key] = value

        # Update the property in-place
        self._property.representations = self._property.representations

    @abstractmethod
    def validate_representation(self, value):
        # type: (Any) -> None
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: Any
        :raises IllegalArgumentError
        :return: None
        """
        pass



predefined_mimes = {
    'image/*': [
        'image/png',
        'image/gif',
        'image/jpeg',
        'image/webp',
        'image/tiff',
        'image/svg+xml',
        'image/bmp'
    ],
    '.xls': [
        'application/vnd.ms-excel',
        'application/x-excel',
        'application/x-msexcel',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    ],
    '.xlsx': [
        'application/vnd.ms-excel',
        'application/x-excel',
        'application/x-msexcel',
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    ],
    'video/*': [
        'video/x-flv',
        'video/mp4',
        'application/x-mpegURL',
        'video/MP2T',
        'video/3gpp',
        'video/quicktime',
        'video/x-msvideo',
        'video/x-ms-wmv',
        'video/mpeg',
        'video/mp1s',
        'video/ogg',
        'video/webm'
    ],
    'model/*': [
        'model/3mf',
        'model/example',
        'model/gltf-binary',
        'model/gltf+json',
        'model/iges',
        'model/stl',
        'model/vnd.collada+xml',
        'model/vnd.dwf',
        'model/vnd.flatland.3dml',
        'model/vnd.gdl',
        'model/vnd.gs-gdl',
        'model/vnd.gtw',
        'model/vnd.moml+xml',
        'model/vnd.mts',
        'model/vnd.opengex',
        'model/vnd.usdz+zip',
        'model/vnd.valve.source.compiled-map',
        'model/vnd.vtu',
        'model/x3d-vrml',
        'model/x3d+fastinfoset',
        'model/x3d+xml'
    ]
}
